import abjad


def test_mutate__set_leaf_duration_01():
    """
    Change leaf to tied duration.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:2])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
            e'8
            f'8
        }
        """
    ), print(abjad.lilypond(voice))

    abjad.mutate._set_leaf_duration(voice[1], abjad.Duration(5, 32))

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ~
            d'32
            ]
            e'8
            f'8
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate__set_leaf_duration_02():
    """
    Change tied leaf to tied value.
    Duplicate ties are not created.
    """

    voice = abjad.Voice("c'8 c'8 c'8 c'8")
    abjad.tie(voice[:2])
    abjad.beam(voice[:2])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            ~
            c'8
            ]
            c'8
            c'8
        }
        """
    ), print(abjad.lilypond(voice))

    abjad.mutate._set_leaf_duration(voice[1], abjad.Duration(5, 32))

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            ~
            c'8
            ~
            c'32
            ]
            c'8
            c'8
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate__set_leaf_duration_03():
    """
    Change leaf to nontied duration.
    Same as voice.written_duration = abjad.Duration(3, 16).
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:2])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
            e'8
            f'8
        }
        """
    ), print(abjad.lilypond(voice))

    abjad.mutate._set_leaf_duration(voice[1], abjad.Duration(3, 16))

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8.
            ]
            e'8
            f'8
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate__set_leaf_duration_04():
    """
    Change leaf to tied duration without power-of-two denominator.
    abjad.Tuplet inserted over new tied notes.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:2])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
            e'8
            f'8
        }
        """
    ), print(abjad.lilypond(voice))

    abjad.mutate._set_leaf_duration(voice[1], abjad.Duration(5, 48))

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            \tweak edge-height #'(0.7 . 0)
            \tuplet 3/2
            {
                d'8
                ~
                d'32
                ]
            }
            e'8
            f'8
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate__set_leaf_duration_05():
    """
    Change leaf to untied duration without power-of-two denominator.
    Tuplet inserted over input leaf.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:2])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
            e'8
            f'8
        }
        """
    ), print(abjad.lilypond(voice))

    abjad.mutate._set_leaf_duration(voice[1], abjad.Duration(1, 12))

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            \tweak edge-height #'(0.7 . 0)
            \tuplet 3/2
            {
                d'8
                ]
            }
            e'8
            f'8
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate__set_leaf_duration_06():
    """
    Change leaf with LilyPond multiplier to untied duration with
    power-of-two denominator. LilyPond multiplier changes but leaf written
    duration does not.
    """

    note = abjad.Note(0, (1, 8), multiplier=(1, 2))

    assert abjad.lilypond(note) == "c'8 * 1/2"

    abjad.mutate._set_leaf_duration(note, abjad.Duration(1, 32))

    assert abjad.wf.wellformed(note)
    assert abjad.lilypond(note) == "c'8 * 1/4"


def test_mutate__set_leaf_duration_07():
    """
    Change leaf with LilyPond multiplier to untied duration with
    power-of-two denominator. LilyPond multiplier changes but leaf
    written duration does not.
    """

    note = abjad.Note(0, (1, 8), multiplier=(1, 2))

    assert abjad.lilypond(note) == "c'8 * 1/2"

    abjad.mutate._set_leaf_duration(note, abjad.Duration(3, 32))

    assert abjad.wf.wellformed(note)
    assert abjad.lilypond(note) == "c'8 * 3/4"


def test_mutate__set_leaf_duration_08():
    """
    Change leaf with LilyPond multiplier to tied duration with
    power-of-two denominator. LilyPond multiplier changes but leaf
    written duration does not.
    """

    note = abjad.Note(0, (1, 8), multiplier=(1, 2))

    assert abjad.lilypond(note) == "c'8 * 1/2"

    abjad.mutate._set_leaf_duration(note, abjad.Duration(5, 32))

    assert abjad.wf.wellformed(note)
    assert abjad.lilypond(note) == "c'8 * 5/4"


def test_mutate__set_leaf_duration_09():
    """
    Change leaf with LilyPond multiplier to duration without
    power-of-two denominator. LilyPond multiplier changes but leaf
    written duration does not.
    """

    note = abjad.Note(0, (1, 8), multiplier=(1, 2))

    assert abjad.lilypond(note) == "c'8 * 1/2"

    abjad.mutate._set_leaf_duration(note, abjad.Duration(1, 24))

    assert abjad.wf.wellformed(note)
    assert abjad.lilypond(note) == "c'8 * 1/3"


def test_mutate__set_leaf_duration_10():
    """
    Change leaf with LilyPond multiplier.
    Change to tie-necessitating duration without power-of-two denominator.
    LilyPond multiplier changes but leaf written duration does not.
    """

    note = abjad.Note(0, (1, 8), multiplier=(1, 2))

    assert abjad.lilypond(note) == "c'8 * 1/2"

    abjad.mutate._set_leaf_duration(note, abjad.Duration(5, 24))

    assert abjad.wf.wellformed(note)
    assert abjad.lilypond(note) == "c'8 * 5/3"


def test_mutate__set_leaf_duration_11():
    """
    Change rest duration.
    """

    voice = abjad.Voice("c'8 r8 e'8 f'8")
    abjad.beam(voice[:3])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            r8
            e'8
            ]
            f'8
        }
        """
    ), print(abjad.lilypond(voice))

    abjad.mutate._set_leaf_duration(voice[1], abjad.Duration(5, 32))

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            r8
            r32
            e'8
            ]
            f'8
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate__split_container_by_duration_01():
    """
    Split one container in score.
    Adds tie after split.
    """

    voice = abjad.Voice()
    voice.append(abjad.Container("c'8 d'8"))
    voice.append(abjad.Container("e'8 f'8"))
    leaves = abjad.select.leaves(voice)
    abjad.beam(leaves[:2])
    abjad.beam(leaves[-2:])
    abjad.slur(leaves)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                (
                d'8
                ]
            }
            {
                e'8
                [
                f'8
                )
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    abjad.mutate._split_container_by_duration(voice[0], abjad.Duration(1, 32))

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'32
                [
                (
                ~
            }
            {
                c'16.
                d'8
                ]
            }
            {
                e'8
                [
                f'8
                )
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate__split_container_by_duration_02():
    """
    Split in-score container at split offset with non-power-of-two denominator.
    """

    voice = abjad.Voice()
    voice.append(abjad.Container("c'8 d'8"))
    voice.append(abjad.Container("e'8 f'8"))
    leaves = abjad.select.leaves(voice)
    abjad.beam(leaves[:2])
    abjad.beam(leaves[-2:])
    abjad.slur(leaves)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                (
                d'8
                ]
            }
            {
                e'8
                [
                f'8
                )
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    abjad.mutate._split_container_by_duration(voice[0], abjad.Duration(1, 5))

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                (
                \tweak edge-height #'(0.7 . 0)
                \tuplet 5/4
                {
                    d'16.
                    ~
                }
            }
            {
                \tweak edge-height #'(0.7 . 0)
                \tuplet 5/4
                {
                    d'16
                    ]
                }
            }
            {
                e'8
                [
                f'8
                )
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate__split_leaf_by_durations_01():
    """
    Splits note into assignable notes.

    Does tie split notes.
    """

    staff = abjad.Staff("c'8 [ d'8 e'8 ]")

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            [
            d'8
            e'8
            ]
        }
        """
    ), print(abjad.lilypond(staff))

    abjad.mutate._split_leaf_by_durations(staff[1], [abjad.Duration(1, 32)])

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            [
            d'32
            ~
            d'16.
            e'8
            ]
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)


def test_mutate__split_leaf_by_durations_02():
    """
    REGRESSION.

    Splits note into tuplet monads and then fuses monads.

    Ties split notes.

    This test comes from #272 in GitHub.
    """

    voice = abjad.Voice(r"\tuplet 3/2 { c'8 [ d'8 e'8 ] }")
    leaf = abjad.get.leaf(voice, 0)
    abjad.mutate._split_leaf_by_durations(leaf, [abjad.Duration(1, 20)])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            \tuplet 3/2
            {
                \tuplet 5/4
                {
                    c'16.
                    [
                    ~
                    c'16
                }
                d'8
                e'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate__split_leaf_by_durations_03():
    """
    Leaf duration less than split duration produces no change.
    """

    staff = abjad.Staff("c'4")
    abjad.mutate._split_leaf_by_durations(staff[0], [abjad.Duration(3, 4)])

    assert len(staff) == 1
    assert isinstance(staff[0], abjad.Note)
    assert staff[0].written_duration == abjad.Duration(1, 4)


def test_mutate__split_leaf_by_durations_04():
    """
    Returns selection of new leaves.
    """

    note = abjad.Note("c'4")
    new_leaves = abjad.mutate._split_leaf_by_durations(note, [abjad.Duration(1, 16)])
    assert all(isinstance(_, abjad.Note) for _ in new_leaves)


def test_mutate__split_leaf_by_durations_05():
    """
    Lone spanned leaf results in two spanned leaves.
    """

    staff = abjad.Staff([abjad.Note("c'4")])
    abjad.mutate._split_leaf_by_durations(staff[0], [abjad.Duration(1, 8)])

    assert len(staff) == 2
    assert abjad.wf.wellformed(staff)


def test_mutate__split_leaf_by_durations_06():
    """
    Returns three leaves with two tied.
    """

    staff = abjad.Staff([abjad.Note("c'4")])
    new_leaves = abjad.mutate._split_leaf_by_durations(
        staff[0], [abjad.Duration(5, 32)]
    )

    assert all(isinstance(_, abjad.Note) for _ in new_leaves)
    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            ~
            c'32
            ~
            c'16.
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)


def test_mutate__split_leaf_by_durations_07():
    """
    After grace notes are removed from first split leaf.
    """

    note = abjad.Note("c'4")
    after_grace = abjad.AfterGraceContainer([abjad.Note(0, (1, 32))])
    abjad.attach(after_grace, note)

    new_leaves = abjad.mutate._split_leaf_by_durations(note, [abjad.Duration(1, 8)])
    staff = abjad.Staff(new_leaves)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            ~
            \afterGrace
            c'8
            {
                c'32
            }
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.get.after_grace_container(new_leaves[0]) is None
    assert len(abjad.get.after_grace_container(new_leaves[1])) == 1
    abjad.wf.wellformed(staff)


def test_mutate__split_leaf_by_durations_08():
    """
    After grace notes are removed from first split leaf.
    """

    note = abjad.Note("c'4")
    grace = abjad.AfterGraceContainer([abjad.Note(0, (1, 32))])
    abjad.attach(grace, note)

    new_leaves = abjad.mutate._split_leaf_by_durations(note, [abjad.Duration(5, 32)])
    staff = abjad.Staff(new_leaves)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            ~
            c'32
            ~
            \afterGrace
            c'16.
            {
                c'32
            }
        }
        """
    ), print(abjad.lilypond(staff))

    abjad.wf.wellformed(staff)


def test_mutate__split_leaf_by_durations_09():
    """
    Grace notes are removed from second split leaf.
    """

    note = abjad.Note("c'4")
    grace = abjad.BeforeGraceContainer([abjad.Note(0, (1, 32))])
    abjad.attach(grace, note)

    new_leaves = abjad.mutate._split_leaf_by_durations(note, [abjad.Duration(1, 16)])
    staff = abjad.Staff(new_leaves)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \grace {
                c'32
            }
            c'16
            ~
            c'8.
        }
        """
    ), print(abjad.lilypond(staff))

    abjad.wf.wellformed(staff)


def test_mutate__split_leaf_by_durations_10():
    """
    Split one leaf in score.
    Ties after split.
    """

    voice = abjad.Voice()
    voice.append(abjad.Container("c'8 d'8"))
    voice.append(abjad.Container("e'8 f'8"))
    leaves = abjad.select.leaves(voice)
    abjad.beam(leaves[:2])
    abjad.beam(leaves[-2:])
    abjad.slur(leaves)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                (
                d'8
                ]
            }
            {
                e'8
                [
                f'8
                )
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    abjad.mutate._split_leaf_by_durations(leaves[0], [abjad.Duration(1, 32)])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'32
                [
                (
                ~
                c'16.
                d'8
                ]
            }
            {
                e'8
                [
                f'8
                )
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate_split_01():
    """
    Cyclically splits note in score.
    """

    voice = abjad.Voice()
    voice.append(abjad.Container("c'8 d'8"))
    voice.append(abjad.Container("e'8 f'8"))
    leaves = abjad.select.leaves(voice)
    abjad.beam(leaves[:2])
    abjad.beam(leaves[-2:])
    abjad.slur(leaves)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                (
                d'8
                ]
            }
            {
                e'8
                [
                f'8
                )
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    notes = voice[0][1:2]
    result = abjad.mutate.split(notes, [abjad.Duration(3, 64)], cyclic=True)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                (
                d'32.
                ~
                d'32.
                ~
                d'32
                ]
            }
            {
                e'8
                [
                f'8
                )
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)
    assert len(result) == 3


def test_mutate_split_02():
    """
    Cyclically splits consecutive notes in score.
    """

    voice = abjad.Voice([abjad.Container("c'8 d'"), abjad.Container("e'8 f'")])
    staff = abjad.Staff([voice])
    abjad.Score([staff], name="Score")
    for container in voice:
        time_signature = abjad.TimeSignature((2, 8))
        abjad.attach(time_signature, container[0])
    leaves = abjad.select.leaves(voice)
    abjad.beam(leaves[:2])
    abjad.beam(leaves[-2:])
    abjad.slur(leaves)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                {
                    \time 2/8
                    c'8
                    [
                    (
                    d'8
                    ]
                }
                {
                    \time 2/8
                    e'8
                    [
                    f'8
                    )
                    ]
                }
            }
        }
        """
    ), print(abjad.lilypond(staff))

    result = abjad.mutate.split(leaves, [abjad.Duration(3, 32)], cyclic=True)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                {
                    \time 2/8
                    c'16.
                    [
                    (
                    ~
                    c'32
                    d'16
                    ~
                    d'16
                    ]
                }
                {
                    \time 2/8
                    e'32
                    [
                    ~
                    e'16.
                    f'16.
                    ~
                    f'32
                    )
                    ]
                }
            }
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)
    assert len(result) == 6


def test_mutate_split_03():
    """
    Cyclically splits note in score.
    """

    voice = abjad.Voice([abjad.Container("c'8 d'"), abjad.Container("e'8 f'")])
    staff = abjad.Staff([voice])
    abjad.Score([staff], name="Score")
    for container in voice:
        time_signature = abjad.TimeSignature((2, 8))
        abjad.attach(time_signature, container[0])
    leaves = abjad.select.leaves(voice)
    abjad.beam(leaves[:2])
    abjad.beam(leaves[-2:])
    abjad.slur(leaves)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                {
                    \time 2/8
                    c'8
                    [
                    (
                    d'8
                    ]
                }
                {
                    \time 2/8
                    e'8
                    [
                    f'8
                    )
                    ]
                }
            }
        }
        """
    ), print(abjad.lilypond(staff))

    notes = voice[0][1:]
    result = abjad.mutate.split(notes, [abjad.Duration(1, 32)], cyclic=True)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                {
                    \time 2/8
                    c'8
                    [
                    (
                    d'32
                    ~
                    d'32
                    ~
                    d'32
                    ~
                    d'32
                    ]
                }
                {
                    \time 2/8
                    e'8
                    [
                    f'8
                    )
                    ]
                }
            }
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)
    assert len(result) == 4


def test_mutate_split_04():
    """
    Cyclically splits consecutive notes in score.
    """

    voice = abjad.Voice([abjad.Container("c'8 d'"), abjad.Container("e'8 f'")])
    staff = abjad.Staff([voice])
    abjad.Score([staff], name="Score")
    for container in voice:
        time_signature = abjad.TimeSignature((2, 8))
        abjad.attach(time_signature, container[0])
    leaves = abjad.select.leaves(voice)
    abjad.beam(leaves[:2])
    abjad.beam(leaves[-2:])
    abjad.slur(leaves)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                {
                    \time 2/8
                    c'8
                    [
                    (
                    d'8
                    ]
                }
                {
                    \time 2/8
                    e'8
                    [
                    f'8
                    )
                    ]
                }
            }
        }
        """
    ), print(abjad.lilypond(staff))

    result = abjad.mutate.split(leaves, [abjad.Duration(1, 16)], cyclic=True)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                {
                    \time 2/8
                    c'16
                    [
                    (
                    ~
                    c'16
                    d'16
                    ~
                    d'16
                    ]
                }
                {
                    \time 2/8
                    e'16
                    [
                    ~
                    e'16
                    f'16
                    ~
                    f'16
                    )
                    ]
                }
            }
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)
    assert len(result) == 8


def test_mutate_split_05():
    """
    Cyclically splits measure in score.
    """

    voice = abjad.Voice([abjad.Container("c'8 d'"), abjad.Container("e'8 f'")])
    staff = abjad.Staff([voice])
    abjad.Score([staff], name="Score")
    for container in voice:
        time_signature = abjad.TimeSignature((2, 8))
        abjad.attach(time_signature, container[0])
    leaves = abjad.select.leaves(voice)
    abjad.beam(leaves[:2])
    abjad.beam(leaves[-2:])
    abjad.slur(leaves)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                {
                    \time 2/8
                    c'8
                    [
                    (
                    d'8
                    ]
                }
                {
                    \time 2/8
                    e'8
                    [
                    f'8
                    )
                    ]
                }
            }
        }
        """
    ), print(abjad.lilypond(staff))

    measures = voice[:1]
    result = abjad.mutate.split(measures, [abjad.Duration(1, 16)], cyclic=True)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                {
                    \time 2/8
                    c'16
                    [
                    (
                    ~
                }
                {
                    c'16
                }
                {
                    d'16
                    ~
                }
                {
                    d'16
                    ]
                }
                {
                    \time 2/8
                    e'8
                    [
                    f'8
                    )
                    ]
                }
            }
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)
    assert len(result) == 4


def test_mutate_split_06():
    """
    Cyclically splits consecutive measures in score.
    """

    voice = abjad.Voice([abjad.Container("c'8 d'"), abjad.Container("e'8 f'")])
    staff = abjad.Staff([voice])
    abjad.Score([staff], name="Score")
    for container in voice:
        time_signature = abjad.TimeSignature((2, 8))
        abjad.attach(time_signature, container[0])
    leaves = abjad.select.leaves(voice)
    abjad.beam(leaves[:2])
    abjad.beam(leaves[-2:])
    abjad.slur(leaves)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                {
                    \time 2/8
                    c'8
                    [
                    (
                    d'8
                    ]
                }
                {
                    \time 2/8
                    e'8
                    [
                    f'8
                    )
                    ]
                }
            }
        }
        """
    ), print(abjad.lilypond(staff))

    measures = voice[:]
    result = abjad.mutate.split(measures, [abjad.Duration(3, 32)], cyclic=True)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                {
                    \time 2/8
                    c'16.
                    [
                    (
                    ~
                }
                {
                    c'32
                    d'16
                    ~
                }
                {
                    d'16
                    ]
                }
                {
                    \time 2/8
                    e'32
                    [
                    ~
                }
                {
                    e'16.
                }
                {
                    f'16.
                    ~
                }
                {
                    f'32
                    )
                    ]
                }
            }
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)
    assert len(result) == 6


def test_mutate_split_07():
    """
    Splits tuplet in score
    """

    voice = abjad.Voice()
    voice.append(abjad.Tuplet((2, 3), "c'8 d'8 e'8"))
    voice.append(abjad.Tuplet((2, 3), "f'8 g'8 a'8"))
    leaves = abjad.select.leaves(voice)
    abjad.beam(leaves)

    tuplets = voice[1:2]
    abjad.mutate.split(tuplets, [abjad.Duration(1, 12)])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            \tuplet 3/2
            {
                c'8
                [
                d'8
                e'8
            }
            \tweak edge-height #'(0.7 . 0)
            \tuplet 3/2
            {
                f'8
            }
            \tweak edge-height #'(0.7 . 0)
            \tuplet 3/2
            {
                g'8
                a'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate_split_08():
    """
    Splits in-score measure with power-of-two denominator.
    """

    voice = abjad.Voice()
    voice.append(abjad.Container("c'8 d'8 e'8"))
    voice.append(abjad.Container("f'8 g'8 a'8"))
    leaves = abjad.select.leaves(voice)
    abjad.beam(leaves)

    measures = voice[1:2]
    abjad.mutate.split(measures, [abjad.Duration(1, 8)])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                d'8
                e'8
            }
            {
                f'8
            }
            {
                g'8
                a'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate_split_09():
    """
    Splits container in middle.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")

    result = abjad.mutate.split([voice], [abjad.Duration(1, 4)])

    assert not len(voice)

    voice_1 = result[0][0]
    voice_2 = result[1][0]

    assert abjad.lilypond(voice_1) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            d'8
        }
        """
    ), print(abjad.lilypond(voice_1))

    assert abjad.wf.wellformed(voice_1)

    assert abjad.lilypond(voice_2) == abjad.string.normalize(
        r"""
        \new Voice
        {
            e'8
            f'8
        }
        """
    ), print(abjad.lilypond(voice_2))

    assert abjad.wf.wellformed(voice_2)


def test_mutate_split_10():
    """
    Splits voice at negative index.
    """

    staff = abjad.Staff([abjad.Voice("c'8 d'8 e'8 f'8")])
    voice = staff[0]

    result = abjad.mutate.split([voice], [abjad.Duration(1, 4)])

    left = result[0][0]
    right = result[1][0]

    assert abjad.lilypond(left) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            d'8
        }
        """
    ), print(abjad.lilypond(left))

    assert abjad.lilypond(right) == abjad.string.normalize(
        r"""
        \new Voice
        {
            e'8
            f'8
        }
        """
    ), print(abjad.lilypond(right))

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                c'8
                d'8
            }
            \new Voice
            {
                e'8
                f'8
            }
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)


def test_mutate_split_11():
    """
    Splits container in score.
    """

    voice = abjad.Voice([abjad.Container("c'8 d'8 e'8 f'8")])
    container = voice[0]
    leaves = abjad.select.leaves(container)
    abjad.beam(leaves)

    result = abjad.mutate.split([container], [abjad.Duration(1, 4)])

    left = result[0][0]
    right = result[1][0]

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                d'8
            }
            {
                e'8
                f'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.lilypond(left) == abjad.string.normalize(
        r"""
        {
            c'8
            [
            d'8
        }
        """
    ), print(abjad.lilypond(left))

    assert abjad.lilypond(right) == abjad.string.normalize(
        r"""
        {
            e'8
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(right))

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
        }
        """
    ), print(abjad.lilypond(container))

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                d'8
            }
            {
                e'8
                f'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate_split_12():
    """
    Splits tuplet in score.
    """

    tuplet = abjad.Tuplet((4, 5), "c'8 c'8 c'8 c'8 c'8")
    voice = abjad.Voice([tuplet])
    staff = abjad.Staff([voice])
    abjad.beam(tuplet[:])

    result = abjad.mutate.split([tuplet], [abjad.Duration(1, 5)])

    left = result[0][0]
    right = result[1][0]

    assert abjad.lilypond(left) == abjad.string.normalize(
        r"""
        \tweak edge-height #'(0.7 . 0)
        \tuplet 5/4
        {
            c'8
            [
            c'8
        }
        """
    ), print(abjad.lilypond(left))

    assert abjad.lilypond(right) == abjad.string.normalize(
        r"""
        \tweak edge-height #'(0.7 . 0)
        \tuplet 5/4
        {
            c'8
            c'8
            c'8
            ]
        }
        """
    ), print(abjad.lilypond(right))

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tuplet 5/4
        {
        }
        """
    ), print(abjad.lilypond(tuplet))

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            \tweak edge-height #'(0.7 . 0)
            \tuplet 5/4
            {
                c'8
                [
                c'8
            }
            \tweak edge-height #'(0.7 . 0)
            \tuplet 5/4
            {
                c'8
                c'8
                c'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                \tweak edge-height #'(0.7 . 0)
                \tuplet 5/4
                {
                    c'8
                    [
                    c'8
                }
                \tweak edge-height #'(0.7 . 0)
                \tuplet 5/4
                {
                    c'8
                    c'8
                    c'8
                    ]
                }
            }
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)


def test_mutate_split_13():
    """
    Splits cyclically.
    """

    voice = abjad.Voice([abjad.Container("c'8 d'8 e'8 f'8 g'8 a'8 b'8 c''8")])
    leaves = abjad.select.leaves(voice)
    abjad.beam(leaves)
    abjad.slur(leaves)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                (
                d'8
                e'8
                f'8
                g'8
                a'8
                b'8
                c''8
                )
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    note = voice[0]
    abjad.mutate.split(note, [abjad.Duration(1, 8), abjad.Duration(3, 8)], cyclic=True)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                (
            }
            {
                d'8
                e'8
                f'8
            }
            {
                g'8
            }
            {
                a'8
                b'8
                c''8
                )
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate_split_14():
    """
    Cyclically splits all components in container.
    """

    voice = abjad.Voice([abjad.Container("c'8 d'8 e'8 f'8")])
    leaves = abjad.select.leaves(voice)
    abjad.beam(leaves)
    abjad.slur(leaves)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                (
                d'8
                e'8
                f'8
                )
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    container = voice[0]
    abjad.mutate.split(container, [abjad.Duration(1, 8)], cyclic=True)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                (
            }
            {
                d'8
            }
            {
                e'8
            }
            {
                f'8
                )
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_mutate_split_15():
    """
    Splits leaf at non-assignable, non-power-of-two offset.
    """

    staff = abjad.Staff("c'4")

    notes = staff[:1]
    abjad.mutate.split(notes, [abjad.Duration(5, 24)])

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \tuplet 3/2
            {
                c'4
                ~
                c'16
                ~
                c'16
            }
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)



import pytest

import abjad


def test_get_timespan_01():
    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    for i, x in enumerate(voice):
        assert abjad.get.timespan(x).start_offset == i * abjad.Offset(1, 8)


def test_get_timespan_02():
    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    for i, x in enumerate(staff):
        assert abjad.get.timespan(x).start_offset == i * abjad.Offset(1, 8)


def test_get_timespan_03():
    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    staff[-1] = abjad.Rest((1, 8))
    for i, x in enumerate(staff):
        assert abjad.get.timespan(x).start_offset == i * abjad.Offset(1, 8)


def test_get_timespan_04():
    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    staff[10:10] = [abjad.Rest((1, 8))]
    for i, x in enumerate(staff):
        assert abjad.get.timespan(x).start_offset == i * abjad.Offset(1, 8)


def test_get_timespan_05():
    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    staff[10:12] = [abjad.Rest((1, 8))]
    for i, x in enumerate(staff):
        assert abjad.get.timespan(x).start_offset == i * abjad.Offset(1, 8)


def test_get_timespan_06():
    """
    Offset works with voices.
    """

    voice_1 = abjad.Voice("c'8 d'8 e'8 f'8")
    voice_2 = abjad.Voice("c'8 d'8 e'8 f'8")
    voice_1.name = voice_2.name = "voice"
    container = abjad.Container([voice_1, voice_2])
    leaves = abjad.select.leaves(container)
    for i, x in enumerate(leaves):
        assert abjad.get.timespan(x).start_offset == i * abjad.Offset(1, 8)


def test_get_timespan_07():
    tuplet = abjad.Tuplet((2, 3), "c'8 c'8 c'8")
    for i, x in enumerate(tuplet):
        assert abjad.get.timespan(x).start_offset == i * abjad.Offset(1, 12)


def test_get_timespan_08():
    tuplet_1 = abjad.Tuplet((2, 3), "c'8 c'8 c'8")
    voice = abjad.Voice([abjad.Note(0, (1, 8)), tuplet_1, abjad.Note(0, (1, 8))])
    offset = 0
    durations = [(1, 8), (1, 12), (1, 12), (1, 12), (1, 8)]
    leaves = abjad.select.leaves(voice)
    for leaf, duration in zip(leaves, durations):
        assert abjad.get.timespan(leaf).start_offset == offset
        offset += abjad.Offset(*duration)


def test_get_timespan_09():
    """
    Offset works on nested tuplets.
    """

    tuplet_1 = abjad.Tuplet((2, 3), "c'8 c'8 c'8")
    tuplet = abjad.Tuplet((2, 3), [abjad.Note("c'4"), tuplet_1, abjad.Note("c'4")])
    offset = 0
    durations = [(1, 6), (1, 18), (1, 18), (1, 18), (1, 6)]
    leaves = abjad.select.leaves(tuplet)
    for leaf, duration in zip(leaves, durations):
        assert abjad.get.timespan(leaf).start_offset == offset
        offset += abjad.Offset(*duration)


def test_get_timespan_10():
    """
    Offset works with simultaneous structures.
    """

    voice_1 = abjad.Voice("c'8 d'8 e'8 f'8")
    voice_2 = abjad.Voice("c'8 d'8 e'8 f'8")
    staff = abjad.Staff([voice_1, voice_2])
    staff.simultaneous = True
    for i, leaf in enumerate(voice_1):
        start_offset = abjad.get.timespan(leaf).start_offset
        assert start_offset == i * abjad.Offset(1, 8)
    for i, leaf in enumerate(voice_2):
        start_offset = abjad.get.timespan(leaf).start_offset
        assert start_offset == i * abjad.Offset(1, 8)


def test_get_timespan_11():
    """
    Offset on leaves works in nested contexts.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    staff = abjad.Staff([abjad.Note(0, (1, 8)), voice, abjad.Note(0, (1, 8))])
    leaves = abjad.select.leaves(staff)
    for i, leaf in enumerate(leaves):
        start_offset = abjad.get.timespan(leaf).start_offset
        assert start_offset == i * abjad.Offset(1, 8)
    leaves = abjad.select.leaves(voice)
    for i, leaf in enumerate(leaves):
        start_offset = abjad.get.timespan(leaf).start_offset
        assert start_offset == i * abjad.Offset(1, 8) + abjad.Offset(1, 8)


def test_get_timespan_12():
    """
    Offset on leaves works in sequential contexts.
    """

    voice_1 = abjad.Voice("c'8 d'8 e'8 f'8")
    voice_2 = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.Staff([voice_1, voice_2])
    for i, leaf in enumerate(voice_1):
        start_offset = abjad.get.timespan(leaf).start_offset
        assert start_offset == i * abjad.Offset(1, 8)
    for i, leaf in enumerate(voice_2):
        start_offset = abjad.get.timespan(leaf).start_offset
        assert start_offset == i * abjad.Offset(1, 8) + abjad.Offset(1, 2)


def test_get_timespan_13():
    """
    Offset on leaves works in nested simultaneous contexts.
    """

    voice_1 = abjad.Voice("c'8 d'8 e'8 f'8")
    voice_2 = abjad.Voice("c'8 d'8 e'8 f'8")
    staff = abjad.Staff([voice_1, voice_2])
    staff.simultaneous = True
    for i, leaf in enumerate(voice_1):
        start_offset = abjad.get.timespan(leaf).start_offset
        assert start_offset == i * abjad.Offset(1, 8)
    for i, leaf in enumerate(voice_2):
        start_offset = abjad.get.timespan(leaf).start_offset
        assert start_offset == i * abjad.Offset(1, 8)


def test_get_timespan_14():
    """
    Offset on leaves works in nested simultaneous and sequential contexts.
    """

    voice_1 = abjad.Voice("c'8 d'8 e'8 f'8")
    voice_2 = abjad.Voice("c'8 d'8 e'8 f'8")
    voice_3 = abjad.Voice("c'8 d'8 e'8 f'8")
    staff = abjad.Staff([abjad.Container([voice_1, voice_2]), voice_3])
    staff[0].simultaneous = True
    for i, leaf in enumerate(voice_3):
        start_offset = abjad.get.timespan(leaf).start_offset
        assert start_offset == i * abjad.Offset(1, 8) + abjad.Offset(1, 2)


def test_get_timespan_15():
    """
    Offset on leaves works in nested simultaneous and sequential contexts.
    """

    voice_1 = abjad.Voice("c'8 d'8 e'8 f'8")
    voice_2 = abjad.Voice("c'8 d'8 e'8 f'8")
    voice_3 = abjad.Voice("c'8 d'8 e'8 f'8")
    staff = abjad.Staff([voice_3, abjad.Container([voice_1, voice_2])])
    staff[1].simultaneous = True
    for i, leaf in enumerate(voice_1):
        start_offset = abjad.get.timespan(leaf).start_offset
        assert start_offset == i * abjad.Offset(1, 8) + abjad.Offset(1, 2)
    for i, leaf in enumerate(voice_2):
        start_offset = abjad.get.timespan(leaf).start_offset
        assert start_offset == i * abjad.Offset(1, 8) + abjad.Offset(1, 2)


def test_get_timespan_16():
    """
    Offsets works on sequential voices.
    """

    staff = abjad.Staff(
        [abjad.Voice("c'8 d'8 e'8 f'8"), abjad.Voice("c'8 d'8 e'8 f'8")]
    )
    staff[0].name = staff[1].name = "voice"
    for i, voice in enumerate(staff):
        start_offset = abjad.get.timespan(voice).start_offset
        assert start_offset == i * abjad.Offset(4, 8)


def test_get_timespan_17():
    """
    Prolated offset does NOT go across sequential staves.
    """

    container = abjad.Container(
        [abjad.Staff("c'8 d'8 e'8 f'8"), abjad.Staff("c'8 d'8 e'8 f'8")]
    )
    container[0].name = container[1].name = "staff"
    start_offset = abjad.get.timespan(container[0]).start_offset
    assert start_offset == abjad.Offset(0)
    start_offset = abjad.get.timespan(container[1]).start_offset
    assert start_offset == abjad.Offset(1, 2)


def test_get_timespan_18():
    """
    Offsets works with nested voices.
    """

    staff = abjad.Staff(
        [abjad.Voice("c'8 d'8 e'8 f'8"), abjad.Voice("c'8 d'8 e'8 f'8")]
    )
    for i, voice in enumerate(staff):
        start_offset = abjad.get.timespan(voice).start_offset
        assert start_offset == i * abjad.Offset(4, 8)


def test_get_timespan_19():
    """
    Offsets works on sequential tuplets.
    """

    voice = abjad.Voice(
        [
            abjad.Tuplet((2, 3), "c'8 d'8 e'8"),
            abjad.Tuplet((2, 3), "c'8 d'8 e'8"),
            abjad.Tuplet((2, 3), "c'8 d'8 e'8"),
        ]
    )
    assert abjad.get.timespan(voice[0]).start_offset == 0 * abjad.Offset(1, 4)
    assert abjad.get.timespan(voice[1]).start_offset == 1 * abjad.Offset(1, 4)
    assert abjad.get.timespan(voice[2]).start_offset == 2 * abjad.Offset(1, 4)


def test_get_timespan_20():
    """
    Offsets work on tuplets between notes.
    """

    tuplet_1 = abjad.Tuplet((2, 3), "c'8 c'8 c'8")
    voice = abjad.Voice([abjad.Note(0, (1, 8)), tuplet_1, abjad.Note(0, (1, 8))])
    assert abjad.get.timespan(voice[0]).start_offset == 0 * abjad.Offset(1, 8)
    assert abjad.get.timespan(voice[1]).start_offset == 1 * abjad.Offset(1, 8)
    assert abjad.get.timespan(voice[2]).start_offset == 3 * abjad.Offset(1, 8)


def test_get_timespan_21():
    """
    Offsets work on nested tuplets.
    """

    tuplet_1 = abjad.Tuplet((1, 2), "c'8 d'8 e'8 f'8")
    contents = [abjad.Note("c'4"), tuplet_1, abjad.Note("c'4")]
    tuplet = abjad.Tuplet((2, 3), contents)
    assert abjad.get.timespan(tuplet[0]).start_offset == 0 * abjad.Offset(1, 6)
    assert abjad.get.timespan(tuplet[1]).start_offset == 1 * abjad.Offset(1, 6)
    assert abjad.get.timespan(tuplet[2]).start_offset == 2 * abjad.Offset(1, 6)


def test_get_timespan_22():
    """
    Offsets work on nested contexts.
    """

    inner_voice = abjad.Voice("c'8 d'8 e'8 f'8")
    outer_voice = abjad.Voice([abjad.Note(0, (1, 8)), inner_voice])
    inner_voice.name = outer_voice.name = "voice"
    abjad.Staff([abjad.Note(1, (1, 8)), outer_voice])
    assert abjad.get.timespan(inner_voice).start_offset == abjad.Offset(2, 8)
    assert abjad.get.timespan(outer_voice).start_offset == abjad.Offset(1, 8)


def test_get_timespan_23():
    """
    Offsets work on nested simultaneous contexts.
    """

    voice_1 = abjad.Voice("c'8 d'8 e'8 f'8")
    voice_2 = abjad.Voice("c'8 d'8 e'8 f'8")
    staff = abjad.Staff([voice_1, voice_2])
    staff.simultaneous = True
    assert abjad.get.timespan(staff[0]).start_offset == 0
    assert abjad.get.timespan(staff[1]).start_offset == 0


def test_get_timespan_24():
    """
    Offsets works in nested simultaneous and sequential contexts.
    """

    voice_1 = abjad.Voice("c'8 d'8 e'8 f'8")
    voice_2 = abjad.Voice("c'8 d'8 e'8 f'8")
    voice_1b = abjad.Voice("c'8 d'8 e'8 f'8")
    voice_2b = abjad.Voice("c'8 d'8 e'8 f'8")
    voice_1.name = voice_1b.name = "voiceOne"
    staff_1 = abjad.Staff([voice_1, voice_1b])
    staff_2 = abjad.Staff([voice_2, voice_2b])
    abjad.StaffGroup([staff_1, staff_2])
    assert abjad.get.timespan(voice_1).start_offset == 0
    assert abjad.get.timespan(voice_2).start_offset == 0
    assert abjad.get.timespan(voice_1b).start_offset == abjad.Offset(4, 8)
    assert abjad.get.timespan(voice_2b).start_offset == abjad.Offset(4, 8)


def test_get_timespan_25():
    """
    Offset seconds can not calculate without excplit metronome mark.
    """

    staff = abjad.Staff("c'8 d'8 e'8 f'8")

    with pytest.raises(Exception):
        abjad.get.timespan(staff[0], in_seconds=True).start_offset


def test_get_timespan_26():
    """
    Offset seconds work with explicit metronome mark.
    """

    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    mark = abjad.MetronomeMark(abjad.Duration(1, 8), 48)
    abjad.attach(mark, staff[0], context="Staff")

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \tempo 8=48
            c'8
            d'8
            e'8
            f'8
        }
        """
    )

    start_offset = abjad.get.timespan(staff[0], in_seconds=True).start_offset
    assert start_offset == abjad.Offset(0)
    start_offset = abjad.get.timespan(staff[1], in_seconds=True).start_offset
    assert start_offset == abjad.Offset(5, 4)


def test_Tuplet_timespan_01():
    staff = abjad.Staff(r"c'4 d'4 \tuplet 3/2 { e'4 f'4 g'4 }")

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'4
            d'4
            \tuplet 3/2
            {
                e'4
                f'4
                g'4
            }
        }
        """
    )

    assert abjad.get.timespan(staff) == abjad.Timespan(0, 1)
    assert abjad.get.timespan(staff[0]) == abjad.Timespan(0, (1, 4))
    assert abjad.get.timespan(staff[1]) == abjad.Timespan((1, 4), (1, 2))
    assert abjad.get.timespan(staff[-1]) == abjad.Timespan((1, 2), 1)



import os

import abjad

configuration = abjad.Configuration()
ly_path = configuration.abjad_directory / "test.ly"
pdf_path = configuration.abjad_directory / "test.pdf"
paths = [ly_path, pdf_path]


def test_persist_as_ly_01():
    """
    Persists LilyPond file when no LilyPond file exists.
    """

    note = abjad.Note("c'4")
    with abjad.FilesystemState(remove=[ly_path]):
        result = abjad.persist.as_ly(note, ly_path)
        assert os.path.isfile(ly_path)
        assert isinstance(result, tuple)


def test_persist_as_ly_02():
    """
    Persists LilyPond file when LilyPond file already exists.
    """

    note = abjad.Note("c'4")
    with abjad.FilesystemState(remove=[ly_path]):
        result = abjad.persist.as_ly(note, ly_path)
        assert isinstance(result, tuple)
        assert os.path.isfile(ly_path)
        abjad.persist.as_ly(note, ly_path)
        assert os.path.isfile(ly_path)


def test_persist_as_pdf_01():
    """
    Persists PDF file when no PDF file exists.
    """

    note = abjad.Note("c'4")
    with abjad.FilesystemState(remove=paths):
        result = abjad.persist.as_pdf(note, pdf_path)
        assert os.path.isfile(pdf_path)
        assert isinstance(result, tuple)


def test_persist_as_pdf_02():
    """
    Persists PDF file when equivalent PDF file already exists.
    """

    note = abjad.Note("c'4")
    with abjad.FilesystemState(remove=paths):
        result = abjad.persist.as_pdf(note, pdf_path)
        assert os.path.isfile(pdf_path)
        assert isinstance(result, tuple)
        os.remove(ly_path)
        abjad.persist.as_pdf(note, pdf_path)
        assert os.path.isfile(pdf_path)



import typing

from . import _iterlib
from . import duration as _duration
from . import pcollections as _pcollections
from . import pitch as _pitch
from . import score as _score
from . import select as _select


@typing.overload
def components(
    argument,
    *,
    exclude=None,
    grace=None,
    reverse=None,
) -> typing.Iterator[_score.Component]:
    pass


@typing.overload
def components(
    argument,
    prototype: typing.Type[_score.Tuplet],
    *,
    exclude=None,
    grace=None,
    reverse=None,
) -> typing.Iterator[_score.Tuplet]:
    pass


@typing.overload
def components(
    argument,
    prototype: typing.Type[_score.Container],
    *,
    exclude=None,
    grace=None,
    reverse=None,
) -> typing.Iterator[_score.Container]:
    pass


@typing.overload
def components(
    argument,
    prototype: typing.Type[_score.Leaf],
    *,
    exclude=None,
    grace=None,
    reverse=None,
) -> typing.Iterator[_score.Leaf]:
    pass


def components(
    argument,
    prototype=None,
    *,
    exclude=None,
    grace=None,
    reverse=None,
):
    r"""
    Iterates components in ``argument``.

    ..  container:: example

        Grace iteration is controlled by a ternary flag.

        >>> music_voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
        >>> container = abjad.BeforeGraceContainer("cs'16")
        >>> abjad.attach(container, music_voice[1])
        >>> obgc = abjad.on_beat_grace_container("g'16 gs' a' as'", music_voice[2:3])
        >>> abjad.attach(abjad.Articulation(">"), obgc[0])
        >>> container = abjad.AfterGraceContainer("fs'16")
        >>> abjad.attach(container, music_voice[3])
        >>> staff = abjad.Staff([music_voice])
        >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \context Voice = "MusicVoice"
                {
                    c'4
                    \grace {
                        cs'16
                    }
                    d'4
                    <<
                        \context Voice = "On_Beat_Grace_Container"
                        {
                            \set fontSize = #-3
                            \slash
                            \voiceOne
                            <
                                \tweak font-size 0
                                \tweak transparent ##t
                                e'
                                g'
                            >16
                            - \accent
                            [
                            (
                            gs'16
                            a'16
                            as'16
                            )
                            ]
                        }
                        \context Voice = "MusicVoice"
                        {
                            \voiceTwo
                            e'4
                        }
                    >>
                    \oneVoice
                    \afterGrace
                    f'4
                    {
                        fs'16
                    }
                }
            }

        Leave ``grace`` unset to iterate grace and nongrace components
        together:

        >>> for component in abjad.iterate.components(staff):
        ...     component
        Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
        Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
        Note("c'4")
        BeforeGraceContainer("cs'16")
        Note("cs'16")
        Note("d'4")
        Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }")
        OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16")
        Chord("<e' g'>16")
        Note("gs'16")
        Note("a'16")
        Note("as'16")
        Voice("e'4", name='MusicVoice')
        Note("e'4")
        Note("f'4")
        AfterGraceContainer("fs'16")
        Note("fs'16")

        >>> for component in abjad.iterate.components(staff, reverse=True):
        ...     component
        Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
        Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
        AfterGraceContainer("fs'16")
        Note("fs'16")
        Note("f'4")
        Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }")
        Voice("e'4", name='MusicVoice')
        Note("e'4")
        OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16")
        Note("as'16")
        Note("a'16")
        Note("gs'16")
        Chord("<e' g'>16")
        Note("d'4")
        BeforeGraceContainer("cs'16")
        Note("cs'16")
        Note("c'4")

        Set ``grace=True`` to iterate only grace components:

        >>> for component in abjad.iterate.components(staff, grace=True):
        ...     component
        BeforeGraceContainer("cs'16")
        Note("cs'16")
        OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16")
        Chord("<e' g'>16")
        Note("gs'16")
        Note("a'16")
        Note("as'16")
        AfterGraceContainer("fs'16")
        Note("fs'16")

        >>> for component in abjad.iterate.components(staff, grace=True, reverse=True):
        ...     component
        AfterGraceContainer("fs'16")
        Note("fs'16")
        OnBeatGraceContainer("<e' g'>16 gs'16 a'16 as'16")
        Note("as'16")
        Note("a'16")
        Note("gs'16")
        Chord("<e' g'>16")
        BeforeGraceContainer("cs'16")
        Note("cs'16")

        Set ``grace=False`` to iterate only nongrace components:

        >>> for component in abjad.iterate.components(staff, grace=False):
        ...     component
        Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
        Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
        Note("c'4")
        Note("d'4")
        Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }")
        Voice("e'4", name='MusicVoice')
        Note("e'4")
        Note("f'4")

        >>> for component in abjad.iterate.components(staff, grace=False, reverse=True):
        ...     component
        Staff("{ c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4 }")
        Voice("c'4 d'4 { { <e' g'>16 gs'16 a'16 as'16 } { e'4 } } f'4", name='MusicVoice')
        Note("f'4")
        Container("{ <e' g'>16 gs'16 a'16 as'16 } { e'4 }")
        Voice("e'4", name='MusicVoice')
        Note("e'4")
        Note("d'4")
        Note("c'4")

    """
    return _iterlib._public_iterate_components(
        argument,
        prototype,
        exclude=exclude,
        grace=grace,
        reverse=reverse,
    )


def leaves(
    argument,
    prototype=None,
    *,
    exclude=None,
    grace=None,
    pitched=None,
    reverse=None,
) -> typing.Iterator:
    r"""
    Iterates leaves in ``argument``.

    ..  container:: example

        Set ``exclude=<annotation>`` to exclude leaves with annotation:

        >>> staff = abjad.Staff()
        >>> score = abjad.Score([staff], name="Score")
        >>> staff.extend("<c' bf'>8 <g' a'>8")
        >>> staff.extend("af'8 r8")
        >>> staff.extend("r8 gf'8")
        >>> abjad.attach(abjad.TimeSignature((2, 8)), staff[0])
        >>> abjad.attach("RED", staff[0])
        >>> abjad.attach("BLUE", staff[1])
        >>> abjad.attach("GREEN", staff[2])
        >>> abjad.attach("RED", staff[3])
        >>> abjad.attach("BLUE", staff[4])
        >>> abjad.attach("GREEN", staff[5])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \time 2/8
                <c' bf'>8
                <g' a'>8
                af'8
                r8
                r8
                gf'8
            }

        >>> for leaf in abjad.iterate.leaves(staff, exclude=['RED', 'BLUE']):
        ...     leaf
        ...
        Note("af'8")
        Note("gf'8")

        Excludes leaves to which ``'RED'`` or ``'BLUE'`` attaches.

    ..  container:: example

        Grace iteration is controlled by a ternary flag.

        >>> music_voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
        >>> container = abjad.BeforeGraceContainer("cs'16")
        >>> abjad.attach(container, music_voice[1])
        >>> obgc = abjad.on_beat_grace_container("g'16 gs' a' as'", music_voice[2:3])
        >>> abjad.attach(abjad.Articulation(">"), obgc[0])
        >>> container = abjad.AfterGraceContainer("fs'16")
        >>> abjad.attach(container, music_voice[3])
        >>> staff = abjad.Staff([music_voice])
        >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \context Voice = "MusicVoice"
                {
                    c'4
                    \grace {
                        cs'16
                    }
                    d'4
                    <<
                        \context Voice = "On_Beat_Grace_Container"
                        {
                            \set fontSize = #-3
                            \slash
                            \voiceOne
                            <
                                \tweak font-size 0
                                \tweak transparent ##t
                                e'
                                g'
                            >16
                            - \accent
                            [
                            (
                            gs'16
                            a'16
                            as'16
                            )
                            ]
                        }
                        \context Voice = "MusicVoice"
                        {
                            \voiceTwo
                            e'4
                        }
                    >>
                    \oneVoice
                    \afterGrace
                    f'4
                    {
                        fs'16
                    }
                }
            }

        Leave ``grace`` unset to iterate grace and nongrace leaves together:

        >>> for leaf in abjad.iterate.leaves(staff):
        ...     leaf
        Note("c'4")
        Note("cs'16")
        Note("d'4")
        Chord("<e' g'>16")
        Note("gs'16")
        Note("a'16")
        Note("as'16")
        Note("e'4")
        Note("f'4")
        Note("fs'16")

        Set ``grace=True`` to iterate only grace leaves:

        >>> for leaf in abjad.iterate.leaves(staff, grace=True):
        ...     leaf
        Note("cs'16")
        Chord("<e' g'>16")
        Note("gs'16")
        Note("a'16")
        Note("as'16")
        Note("fs'16")

        Set ``grace=False`` to iterate only nongrace leaves:

        >>> for leaf in abjad.iterate.leaves(staff, grace=False):
        ...     leaf
        Note("c'4")
        Note("d'4")
        Note("e'4")
        Note("f'4")

    ..  container:: example

        Pitched iteration is controlled by a ternary flag.

        >>> staff = abjad.Staff()
        >>> score = abjad.Score([staff], name="Score")
        >>> staff.extend("<c' bf'>8 <g' a'>8")
        >>> staff.extend("af'8 r8")
        >>> staff.extend("r8 gf'8")
        >>> abjad.attach(abjad.TimeSignature((2, 8)), staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \time 2/8
                <c' bf'>8
                <g' a'>8
                af'8
                r8
                r8
                gf'8
            }

        Leaves ``pitched`` unset to iterate pitched and unpitched leaves
        together:

        >>> for leaf in abjad.iterate.leaves(staff):
        ...     leaf
        ...
        Chord("<c' bf'>8")
        Chord("<g' a'>8")
        Note("af'8")
        Rest('r8')
        Rest('r8')
        Note("gf'8")

        >>> for leaf in abjad.iterate.leaves(staff, reverse=True):
        ...     leaf
        Note("gf'8")
        Rest('r8')
        Rest('r8')
        Note("af'8")
        Chord("<g' a'>8")
        Chord("<c' bf'>8")

        Set ``pitched=True`` to iterate pitched leaves only:

        >>> for leaf in abjad.iterate.leaves(staff, pitched=True):
        ...     leaf
        ...
        Chord("<c' bf'>8")
        Chord("<g' a'>8")
        Note("af'8")
        Note("gf'8")

        >>> for leaf in abjad.iterate.leaves(staff, pitched=True, reverse=True):
        ...     leaf
        Note("gf'8")
        Note("af'8")
        Chord("<g' a'>8")
        Chord("<c' bf'>8")

        Set ``pitched=False`` to iterate unpitched leaves only:

        >>> for leaf in abjad.iterate.leaves(staff, pitched=False):
        ...     leaf
        ...
        Rest('r8')
        Rest('r8')

        >>> for leaf in abjad.iterate.leaves(staff, pitched=False):
        ...     leaf
        ...
        Rest('r8')
        Rest('r8')

    """
    return _iterlib._public_iterate_leaves(
        argument,
        prototype=prototype,
        exclude=exclude,
        grace=grace,
        pitched=pitched,
        reverse=reverse,
    )


def logical_ties(
    argument,
    *,
    exclude=None,
    grace=None,
    nontrivial=None,
    pitched=None,
    reverse=None,
) -> typing.Iterator[_select.LogicalTie]:
    r"""
    Iterates logical ties in ``argument``.

    ..  container:: example

        Iterates logical ties:

        >>> string = r"c'4 ~ \times 2/3 { c'16 d'8 } e'8 f'4 ~ f'16"
        >>> staff = abjad.Staff(string)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                ~
                \tuplet 3/2
                {
                    c'16
                    d'8
                }
                e'8
                f'4
                ~
                f'16
            }

        >>> for logical_tie in abjad.iterate.logical_ties(staff):
        ...     logical_tie
        ...
        LogicalTie(items=[Note("c'4"), Note("c'16")])
        LogicalTie(items=[Note("d'8")])
        LogicalTie(items=[Note("e'8")])
        LogicalTie(items=[Note("f'4"), Note("f'16")])

    ..  container:: example

        Grace iteration is controlled by a ternary flag.

        >>> music_voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
        >>> container = abjad.BeforeGraceContainer("cs'16")
        >>> abjad.attach(container, music_voice[1])
        >>> obgc = abjad.on_beat_grace_container("g'16 gs' a' as'", music_voice[2:3])
        >>> abjad.attach(abjad.Articulation(">"), obgc[0])
        >>> container = abjad.AfterGraceContainer("fs'16")
        >>> abjad.attach(container, music_voice[3])
        >>> staff = abjad.Staff([music_voice])
        >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \context Voice = "MusicVoice"
                {
                    c'4
                    \grace {
                        cs'16
                    }
                    d'4
                    <<
                        \context Voice = "On_Beat_Grace_Container"
                        {
                            \set fontSize = #-3
                            \slash
                            \voiceOne
                            <
                                \tweak font-size 0
                                \tweak transparent ##t
                                e'
                                g'
                            >16
                            - \accent
                            [
                            (
                            gs'16
                            a'16
                            as'16
                            )
                            ]
                        }
                        \context Voice = "MusicVoice"
                        {
                            \voiceTwo
                            e'4
                        }
                    >>
                    \oneVoice
                    \afterGrace
                    f'4
                    {
                        fs'16
                    }
                }
            }

        Leave ``grace`` unset to iterate grace and nongrace logical ties together:

        >>> for lt in abjad.iterate.logical_ties(staff):
        ...     lt
        LogicalTie(items=[Note("c'4")])
        LogicalTie(items=[Note("cs'16")])
        LogicalTie(items=[Note("d'4")])
        LogicalTie(items=[Chord("<e' g'>16")])
        LogicalTie(items=[Note("gs'16")])
        LogicalTie(items=[Note("a'16")])
        LogicalTie(items=[Note("as'16")])
        LogicalTie(items=[Note("e'4")])
        LogicalTie(items=[Note("f'4")])
        LogicalTie(items=[Note("fs'16")])

        >>> for lt in abjad.iterate.logical_ties(staff, reverse=True):
        ...     lt
        LogicalTie(items=[Note("fs'16")])
        LogicalTie(items=[Note("f'4")])
        LogicalTie(items=[Note("e'4")])
        LogicalTie(items=[Note("as'16")])
        LogicalTie(items=[Note("a'16")])
        LogicalTie(items=[Note("gs'16")])
        LogicalTie(items=[Chord("<e' g'>16")])
        LogicalTie(items=[Note("d'4")])
        LogicalTie(items=[Note("cs'16")])
        LogicalTie(items=[Note("c'4")])

        Set ``grace=True`` to iterate grace logical ties only:

        >>> for lt in abjad.iterate.logical_ties(staff, grace=True):
        ...     lt
        LogicalTie(items=[Note("cs'16")])
        LogicalTie(items=[Chord("<e' g'>16")])
        LogicalTie(items=[Note("gs'16")])
        LogicalTie(items=[Note("a'16")])
        LogicalTie(items=[Note("as'16")])
        LogicalTie(items=[Note("fs'16")])

        >>> for lt in abjad.iterate.logical_ties(staff, grace=True, reverse=True):
        ...     lt
        LogicalTie(items=[Note("fs'16")])
        LogicalTie(items=[Note("as'16")])
        LogicalTie(items=[Note("a'16")])
        LogicalTie(items=[Note("gs'16")])
        LogicalTie(items=[Chord("<e' g'>16")])
        LogicalTie(items=[Note("cs'16")])

        Set ``grace=False`` to iterate nongrace logical ties only:

        >>> for lt in abjad.iterate.logical_ties(staff, grace=False):
        ...     lt
        LogicalTie(items=[Note("c'4")])
        LogicalTie(items=[Note("d'4")])
        LogicalTie(items=[Note("e'4")])
        LogicalTie(items=[Note("f'4")])

        >>> for lt in abjad.iterate.logical_ties(staff, grace=False, reverse=True):
        ...     lt
        LogicalTie(items=[Note("f'4")])
        LogicalTie(items=[Note("e'4")])
        LogicalTie(items=[Note("d'4")])
        LogicalTie(items=[Note("c'4")])

    ..  container:: example

        Logical tie triviality is controlled by a ternary flag.

        >>> string = r"c'4 ~ \times 2/3 { c'8 d'4 }"
        >>> string += r" e'4 ~ \times 2/3 { e'8 f' }"
        >>> staff = abjad.Staff(string)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                ~
                \tuplet 3/2
                {
                    c'8
                    d'4
                }
                e'4
                ~
                \tweak edge-height #'(0.7 . 0)
                \tuplet 3/2
                {
                    e'8
                    f'8
                }
            }

        Leave ``nontrivial`` unset to iterate trivial and nontrivial
        logical ties together:

        >>> for lt in abjad.iterate.logical_ties(staff):
        ...     lt
        LogicalTie(items=[Note("c'4"), Note("c'8")])
        LogicalTie(items=[Note("d'4")])
        LogicalTie(items=[Note("e'4"), Note("e'8")])
        LogicalTie(items=[Note("f'8")])

        >>> for lt in abjad.iterate.logical_ties(staff, reverse=True):
        ...     lt
        LogicalTie(items=[Note("f'8")])
        LogicalTie(items=[Note("e'4"), Note("e'8")])
        LogicalTie(items=[Note("d'4")])
        LogicalTie(items=[Note("c'4"), Note("c'8")])

        Set ``nontrivial=True`` to iterate nontrivial logical ties only:

        >>> for lt in abjad.iterate.logical_ties(staff, nontrivial=True):
        ...     lt
        LogicalTie(items=[Note("c'4"), Note("c'8")])
        LogicalTie(items=[Note("e'4"), Note("e'8")])

        >>> for lt in abjad.iterate.logical_ties(staff, nontrivial=True, reverse=True):
        ...     lt
        LogicalTie(items=[Note("e'4"), Note("e'8")])
        LogicalTie(items=[Note("c'4"), Note("c'8")])

        Set ``nontrivial=False`` to iterate trivial logical ties only:

        >>> for lt in abjad.iterate.logical_ties(staff, nontrivial=False):
        ...     lt
        LogicalTie(items=[Note("d'4")])
        LogicalTie(items=[Note("f'8")])

        >>> for lt in abjad.iterate.logical_ties(staff, nontrivial=False, reverse=True):
        ...     lt
        LogicalTie(items=[Note("f'8")])
        LogicalTie(items=[Note("d'4")])

    ..  container:: example

        Logical tie pitchedness is controlled by a ternary flag.

        >>> string = r"c'4 ~ \times 2/3 { c'8 r4 }"
        >>> string += r"d'4 ~ \times 2/3 { d'8 r4 }"
        >>> staff = abjad.Staff(string)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                ~
                \tuplet 3/2
                {
                    c'8
                    r4
                }
                d'4
                ~
                \tuplet 3/2
                {
                    d'8
                    r4
                }
            }

        Leave ``pitched`` unset to iterate pitched and unpitched logical ties together:

        >>> for lt in abjad.iterate.logical_ties(staff):
        ...     lt
        LogicalTie(items=[Note("c'4"), Note("c'8")])
        LogicalTie(items=[Rest('r4')])
        LogicalTie(items=[Note("d'4"), Note("d'8")])
        LogicalTie(items=[Rest('r4')])

        >>> for lt in abjad.iterate.logical_ties(staff, reverse=True):
        ...     lt
        LogicalTie(items=[Rest('r4')])
        LogicalTie(items=[Note("d'4"), Note("d'8")])
        LogicalTie(items=[Rest('r4')])
        LogicalTie(items=[Note("c'4"), Note("c'8")])

        Set ``pitched=True`` to iterate pitched logical ties only:

        >>> for lt in abjad.iterate.logical_ties(staff, pitched=True):
        ...     lt
        LogicalTie(items=[Note("c'4"), Note("c'8")])
        LogicalTie(items=[Note("d'4"), Note("d'8")])

        >>> for lt in abjad.iterate.logical_ties(staff, pitched=True, reverse=True):
        ...     lt
        LogicalTie(items=[Note("d'4"), Note("d'8")])
        LogicalTie(items=[Note("c'4"), Note("c'8")])

        Set ``pitched=False`` to iterate unpitched logical ties only:

        >>> for lt in abjad.iterate.logical_ties(staff, pitched=False):
        ...     lt
        LogicalTie(items=[Rest('r4')])
        LogicalTie(items=[Rest('r4')])

        >>> for lt in abjad.iterate.logical_ties(staff, pitched=False, reverse=True):
        ...     lt
        LogicalTie(items=[Rest('r4')])
        LogicalTie(items=[Rest('r4')])

    ..  container:: example

        REGRESSION. Yields logical tie even when leaves are missing in
        input:

        >>> voice = abjad.Voice("c'8 [ ~ c' ~ c' d' ]")
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'8
                [
                ~
                c'8
                ~
                c'8
                d'8
                ]
            }

        >>> selection = voice[:2]
        >>> for logical_tie in abjad.iterate.logical_ties(selection):
        ...     logical_tie
        ...
        LogicalTie(items=[Note("c'8"), Note("c'8"), Note("c'8")])

    """
    return _iterlib._iterate_logical_ties(
        argument,
        exclude=exclude,
        grace=grace,
        nontrivial=nontrivial,
        pitched=pitched,
        reverse=reverse,
    )


def pitches(argument) -> typing.Iterator[_pitch.NamedPitch]:
    r"""
    Iterates pitches in ``argument``.

    ..  container:: example

        Iterates pitches in container:

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'8
                d'8
                e'8
                f'8
            }

        >>> for pitch in abjad.iterate.pitches(staff):
        ...     pitch
        ...
        NamedPitch("c'")
        NamedPitch("d'")
        NamedPitch("e'")
        NamedPitch("f'")

    ..  container:: example

        Iterates pitches in pitch set:

        >>> pitch_set = abjad.PitchSet([0, 2, 4, 5])

        >>> for pitch in abjad.iterate.pitches(pitch_set):
        ...     pitch
        ...
        NumberedPitch(0)
        NumberedPitch(2)
        NumberedPitch(4)
        NumberedPitch(5)

    ..  container:: example

        Iterates different types of object in tuple:

        >>> argument = (
        ...     abjad.NamedPitch("c'"),
        ...     abjad.Note("d'4"),
        ...     abjad.Chord("<e' fs' g>4"),
        ...     )

        >>> for pitch in abjad.iterate.pitches(argument):
        ...     pitch
        ...
        NamedPitch("c'")
        NamedPitch("d'")
        NamedPitch('g')
        NamedPitch("e'")
        NamedPitch("fs'")

    """
    if isinstance(argument, _pitch.Pitch):
        pitch = _pitch.NamedPitch(argument)
        yield pitch
    result = []
    try:
        result.extend(argument.pitches)
    except AttributeError:
        pass
    if isinstance(argument, _score.Chord):
        result.extend(argument.written_pitches)
    elif isinstance(argument, _pcollections.PitchSet):
        result.extend(sorted(list(argument)))
    elif isinstance(argument, list | tuple | set):
        for item in argument:
            for pitch_ in pitches(item):
                result.append(pitch_)
    else:
        for leaf in leaves(argument):
            try:
                result.append(leaf.written_pitch)
            except AttributeError:
                pass
            try:
                result.extend(leaf.written_pitches)
            except AttributeError:
                pass
    for pitch in result:
        yield pitch


def timeline(
    argument, prototype=None, *, exclude=None, reverse=None
) -> tuple[_score.Component, ...]:
    r"""
    Iterates leaves in ``argument`` in timeline order.

    ..  container:: example

        Timeline-iterates leaves:

        >>> score = abjad.Score()
        >>> score.append(abjad.Staff("c'4 d'4 e'4 f'4"))
        >>> score.append(abjad.Staff("g'8 a'8 b'8 c''8"))
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(score)
            >>> print(string)
            \new Score
            <<
                \new Staff
                {
                    c'4
                    d'4
                    e'4
                    f'4
                }
                \new Staff
                {
                    g'8
                    a'8
                    b'8
                    c''8
                }
            >>

        >>> for leaf in abjad.iterate.timeline(score):
        ...     leaf
        ...
        Note("c'4")
        Note("g'8")
        Note("a'8")
        Note("d'4")
        Note("b'8")
        Note("c''8")
        Note("e'4")
        Note("f'4")

        >>> for component in abjad.iterate.timeline(score, reverse=True):
        ...     component
        ...
        Note("f'4")
        Note("e'4")
        Note("c''8")
        Note("b'8")
        Note("d'4")
        Note("a'8")
        Note("g'8")
        Note("c'4")

    ..  container:: example

        REGRESSION. Works with grace note (and containers):

        >>> music_voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
        >>> container = abjad.BeforeGraceContainer("cs'16")
        >>> abjad.attach(container, music_voice[1])
        >>> obgc = abjad.on_beat_grace_container("g'16 gs' a' as'", music_voice[2:3])
        >>> abjad.attach(abjad.Articulation(">"), obgc[0])
        >>> container = abjad.AfterGraceContainer("fs'16")
        >>> abjad.attach(container, music_voice[3])
        >>> staff = abjad.Staff([music_voice])
        >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \context Voice = "MusicVoice"
                {
                    c'4
                    \grace {
                        cs'16
                    }
                    d'4
                    <<
                        \context Voice = "On_Beat_Grace_Container"
                        {
                            \set fontSize = #-3
                            \slash
                            \voiceOne
                            <
                                \tweak font-size 0
                                \tweak transparent ##t
                                e'
                                g'
                            >16
                            - \accent
                            [
                            (
                            gs'16
                            a'16
                            as'16
                            )
                            ]
                        }
                        \context Voice = "MusicVoice"
                        {
                            \voiceTwo
                            e'4
                        }
                    >>
                    \oneVoice
                    \afterGrace
                    f'4
                    {
                        fs'16
                    }
                }
            }

        >>> for leaf in abjad.iterate.timeline(staff):
        ...     leaf
        Note("c'4")
        Note("cs'16")
        Note("d'4")
        Chord("<e' g'>16")
        Note("gs'16")
        Note("a'16")
        Note("as'16")
        Note("e'4")
        Note("f'4")
        Note("fs'16")

        >>> for leaf in abjad.iterate.timeline(staff, reverse=True):
        ...     leaf
        Note("fs'16")
        Note("f'4")
        Note("e'4")
        Note("as'16")
        Note("a'16")
        Note("gs'16")
        Chord("<e' g'>16")
        Note("d'4")
        Note("cs'16")
        Note("c'4")

    Iterates leaves when ``prototype`` is none.
    """
    generator = leaves(argument, prototype=prototype, exclude=exclude)
    components = list(generator)
    components.sort(key=lambda _: _._get_timespan().start_offset)
    offset_to_components: dict[_duration.Offset, list[_score.Component]] = dict()
    for component in components:
        start_offset = component._get_timespan().start_offset
        if start_offset not in offset_to_components:
            offset_to_components[start_offset] = []
    for component in components:
        start_offset = component._get_timespan().start_offset
        offset_to_components[start_offset].append(component)
    result: list[_score.Component] = []
    for start_offset, list_ in offset_to_components.items():
        result.extend(list_)
    if reverse:
        result.reverse()
    return tuple(result)



import dataclasses
import typing

from . import _getlib
from . import bind as _bind
from . import get as _get
from . import indicators as _indicators
from . import instruments as _instruments
from . import iterate as _iterate
from . import pcollections as _pcollections
from . import pitch as _pitch
from . import score as _score


def iterate_out_of_range(argument) -> typing.Iterator[_score.Leaf]:
    r"""
    Iterates out-of-range notes and chords in ``argument``.

    ..  container:: example

        >>> staff = abjad.Staff("c'8 r8 <d fs>8 r8")
        >>> violin = abjad.Violin()
        >>> abjad.attach(violin, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'8
                r8
                <d fs>8
                r8
            }

        >>> for leaf in abjad.iterpitches.iterate_out_of_range(staff):
        ...     leaf
        ...
        Chord('<d fs>8')

    """
    for leaf in _iterate.leaves(argument, pitched=True):
        instrument = _getlib._get_effective(leaf, _instruments.Instrument)
        if instrument is None:
            raise ValueError("no instrument found.")
        if not sounding_pitches_are_in_range(leaf, instrument.pitch_range):
            yield leaf


def respell_with_flats(argument) -> None:
    r"""
    Respells pitches in ``argument`` with flats.

    ..  container:: example

        >>> staff = abjad.Staff("cs'8 ds' es' fs' gs' as' bs'4")
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                cs'8
                ds'8
                es'8
                fs'8
                gs'8
                as'8
                bs'4
            }

        >>> abjad.iterpitches.respell_with_flats(staff)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                df'8
                ef'8
                f'8
                gf'8
                af'8
                bf'8
                c''4
            }

    """
    for leaf in _iterate.leaves(argument):
        if isinstance(leaf, _score.Note):
            assert leaf.written_pitch is not None
            leaf.written_pitch = leaf.written_pitch.respell(accidental="flats")
        elif isinstance(leaf, _score.Chord):
            for note_head in leaf.note_heads:
                pitch = note_head.written_pitch.respell(accidental="flats")
                note_head.written_pitch = pitch


def respell_with_sharps(argument) -> None:
    r"""
    Respells pitches in ``argument`` with sharps.

    ..  container:: example

        >>> staff = abjad.Staff("cf'8 df' ef' ff' gf' af' bf'4")
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                cf'8
                df'8
                ef'8
                ff'8
                gf'8
                af'8
                bf'4
            }

        >>> abjad.iterpitches.respell_with_sharps(staff)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                b8
                cs'8
                ds'8
                e'8
                fs'8
                gs'8
                as'4
            }

    """
    for leaf in _iterate.leaves(argument):
        if isinstance(leaf, _score.Note):
            assert leaf.written_pitch is not None
            leaf.written_pitch = leaf.written_pitch.respell(accidental="sharps")
        elif isinstance(leaf, _score.Chord):
            for note_head in leaf.note_heads:
                pitch = note_head.written_pitch.respell(accidental="sharps")
                note_head.written_pitch = pitch


def sounding_pitches_are_in_range(argument, pitch_range) -> bool:
    """
    Returns true when all pitches in ``argument`` sound within ``pitch_range``.
    """
    assert isinstance(pitch_range, _pcollections.PitchRange), repr(pitch_range)
    if isinstance(argument, int | float):
        pitch = _pitch.NamedPitch(argument)
        return pitch in pitch_range
    if isinstance(argument, _pitch.Pitch):
        return argument in pitch_range
    if hasattr(argument, "written_pitch"):
        sounding_pitch = _getlib._get_sounding_pitch(argument)
        return sounding_pitch in pitch_range
    if hasattr(argument, "written_pitches"):
        sounding_pitches = _getlib._get_sounding_pitches(argument)
        return all(_ in pitch_range for _ in sounding_pitches)
    pitches = list(_iterate.pitches(argument))
    if pitches:
        return all(_ in pitch_range for _ in pitches)
    else:
        try:
            return all(_ in pitch_range for _ in argument)
        except TypeError:
            return False
    return False


def transpose_from_sounding_pitch(argument) -> None:
    r"""
    Transpose notes and chords in ``argument`` from sounding pitch to written pitch.

    ..  container:: example

        >>> staff = abjad.Staff("<c' e' g'>4 d'4 r4 e'4")
        >>> clarinet = abjad.ClarinetInBFlat()
        >>> abjad.attach(clarinet, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                <c' e' g'>4
                d'4
                r4
                e'4
            }

        >>> abjad.iterpitches.transpose_from_sounding_pitch(staff)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                <d' fs' a'>4
                e'4
                r4
                fs'4
            }

    """
    for leaf in _iterate.leaves(argument, pitched=True):
        instrument = _getlib._get_effective(leaf, _instruments.Instrument)
        if not instrument:
            continue
        sounding_pitch = instrument.middle_c_sounding_pitch
        interval = _pitch.NamedPitch("C4") - sounding_pitch
        interval *= -1
        if hasattr(leaf, "note_head"):
            pitch = leaf.written_pitch
            pitch = interval.transpose(pitch)
            leaf.written_pitch = pitch
        elif hasattr(leaf, "note_heads"):
            pitches = [interval.transpose(pitch) for pitch in leaf.written_pitches]
            for note_head, pitch in zip(leaf.note_heads, pitches, strict=True):
                note_head.written_pitch = pitch
        wrapper = _get.indicator(leaf, _indicators.StartTrillSpan, unwrap=False)
        if wrapper is not None:
            start_trill_span = wrapper.unbundle_indicator()
            new_pitch = interval.transpose(start_trill_span.pitch)
            new_start_trill_span = dataclasses.replace(
                start_trill_span, pitch=new_pitch
            )
            wrapper_tag = wrapper.tag
            _bind.detach(wrapper, leaf)
            if wrapper.bundled():
                new_bundle = dataclasses.replace(
                    wrapper.get_item(), indicator=new_start_trill_span
                )
                _bind.attach(new_bundle, leaf, tag=wrapper_tag)
            else:
                _bind.attach(new_start_trill_span, leaf, tag=wrapper_tag)


def transpose_from_written_pitch(argument) -> None:
    r"""
    Transposes notes and chords in ``argument`` from sounding pitch to written pitch.

    ..  container:: example

        >>> staff = abjad.Staff("<c' e' g'>4 d'4 r4 e'4")
        >>> clarinet = abjad.ClarinetInBFlat()
        >>> abjad.attach(clarinet, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                <c' e' g'>4
                d'4
                r4
                e'4
            }

        >>> abjad.iterpitches.transpose_from_written_pitch(staff)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                <bf d' f'>4
                c'4
                r4
                d'4
            }

    """
    for leaf in _iterate.leaves(argument, pitched=True):
        instrument = _getlib._get_effective(leaf, _instruments.Instrument)
        if not instrument:
            continue
        sounding_pitch = instrument.middle_c_sounding_pitch
        interval = _pitch.NamedPitch("C4") - sounding_pitch
        if hasattr(leaf, "note_head"):
            written_pitch = leaf.written_pitch
            written_pitch = interval.transpose(written_pitch)
            leaf.written_pitch = written_pitch
        elif hasattr(leaf, "note_heads"):
            pitches = [interval.transpose(pitch) for pitch in leaf.written_pitches]
            for note_head, pitch in zip(leaf.note_heads, pitches, strict=True):
                note_head.written_pitch = pitch
        wrapper = _get.indicator(leaf, _indicators.StartTrillSpan, unwrap=False)
        if wrapper is not None:
            start_trill_span = wrapper.unbundle_indicator()
            new_pitch = interval.transpose(start_trill_span.pitch)
            new_start_trill_span = dataclasses.replace(
                start_trill_span, pitch=new_pitch
            )
            _bind.detach(wrapper, leaf)
            if wrapper.bundled():
                new_bundle = dataclasses.replace(
                    wrapper.get_item, indicator=new_start_trill_span
                )
                _bind.attach(new_bundle, leaf)
            else:
                _bind.attach(new_start_trill_span, leaf)



r"""
Instruments.

..  container:: example

    Default ranges of each instrument:

    >>> names = [_ for _ in dir(abjad.instruments) if _[0].isupper()]
    >>> for name in names:
    ...     class_ = abjad.__dict__[name]
    ...     if issubclass(class_, abjad.Instrument):
    ...         instrument = class_()
    ...         instrument_name = instrument.__class__.__name__
    ...         range_string = instrument.pitch_range.range_string
    ...         print(f"{instrument_name}: {range_string}")
    Accordion: [E1, C8]
    AltoFlute: [G3, G6]
    AltoSaxophone: [Db3, A5]
    AltoTrombone: [A2, Bb5]
    AltoVoice: [F3, G5]
    BaritoneSaxophone: [C2, Ab4]
    BaritoneVoice: [A2, A4]
    BassClarinet: [Bb1, G5]
    BassFlute: [C3, C6]
    BassSaxophone: [Ab2, E4]
    BassTrombone: [C2, F4]
    BassVoice: [E2, F4]
    Bassoon: [Bb1, Eb5]
    Cello: [C2, G5]
    ClarinetInA: [Db3, A6]
    ClarinetInBFlat: [D3, Bb6]
    ClarinetInEFlat: [F3, C7]
    Contrabass: [C1, G4]
    ContrabassClarinet: [Bb0, G4]
    ContrabassFlute: [G2, G5]
    ContrabassSaxophone: [C1, Ab3]
    Contrabassoon: [Bb0, Bb4]
    EnglishHorn: [E3, C6]
    Flute: [C4, D7]
    FrenchHorn: [B1, F5]
    Glockenspiel: [G5, C8]
    Guitar: [E2, E5]
    Harp: [B0, G#7]
    Harpsichord: [C2, C7]
    Instrument: [-inf, +inf]
    Marimba: [F2, C7]
    MezzoSopranoVoice: [A3, C6]
    Oboe: [Bb3, A6]
    Percussion: [-inf, +inf]
    Piano: [A0, C8]
    Piccolo: [D5, C8]
    SopraninoSaxophone: [Db4, F#6]
    SopranoSaxophone: [Ab3, E6]
    SopranoVoice: [C4, E6]
    TenorSaxophone: [Ab2, E5]
    TenorTrombone: [E2, Eb5]
    TenorVoice: [C3, D5]
    Trumpet: [F#3, D6]
    Tuba: [D1, F4]
    Vibraphone: [F3, F6]
    Viola: [C3, D6]
    Violin: [G3, G7]
    Xylophone: [C4, C7]

..  container:: example

    Two instruments active on a single staff:

    >>> voice_1 = abjad.Voice("e'8 g'8 f'8 a'8")
    >>> voice_2 = abjad.Voice("c'2")
    >>> staff = abjad.Staff([voice_1, voice_2], simultaneous=True)
    >>> flute = abjad.Flute()
    >>> abjad.attach(flute, voice_1[0], context="Voice")
    >>> abjad.attach(abjad.VoiceNumber(1), voice_1[0])
    >>> abjad.attach(abjad.VoiceNumber(2), voice_2[0])
    >>> viola = abjad.Viola()
    >>> abjad.attach(viola, voice_2[0], context="Voice")
    >>> abjad.show(staff) # doctest: +SKIP

    ..  docs::

        >>> string = abjad.lilypond(staff)
        >>> print(string)
        \new Staff
        <<
            \new Voice
            {
                \voiceOne
                e'8
                g'8
                f'8
                a'8
            }
            \new Voice
            {
                \voiceTwo
                c'2
            }
        >>

    >>> for leaf in abjad.select.leaves(voice_1):
    ...     leaf, abjad.get.effective(leaf, abjad.Instrument)
    ...
    (Note("e'8"), Flute(clefs=('treble',), context='Staff', middle_c_sounding_pitch=NamedPitch("c'"), pitch_range=PitchRange(range_string='[C4, D7]')))
    (Note("g'8"), Flute(clefs=('treble',), context='Staff', middle_c_sounding_pitch=NamedPitch("c'"), pitch_range=PitchRange(range_string='[C4, D7]')))
    (Note("f'8"), Flute(clefs=('treble',), context='Staff', middle_c_sounding_pitch=NamedPitch("c'"), pitch_range=PitchRange(range_string='[C4, D7]')))
    (Note("a'8"), Flute(clefs=('treble',), context='Staff', middle_c_sounding_pitch=NamedPitch("c'"), pitch_range=PitchRange(range_string='[C4, D7]')))

    >>> for leaf in abjad.select.leaves(voice_2):
    ...     leaf, abjad.get.effective(leaf, abjad.Instrument)
    ...
    (Note("c'2"), Viola(clefs=('alto', 'treble'), context='Staff', middle_c_sounding_pitch=NamedPitch("c'"), pitch_range=PitchRange(range_string='[C3, D6]'), tuning=Tuning(pitches=(NamedPitch('c'), NamedPitch('g'), NamedPitch("d'"), NamedPitch("a'")))))

"""

import dataclasses
import typing

from . import contributions as _contributions
from . import enumerate as _enumerate
from . import pcollections as _pcollections
from . import pitch as _pitch


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Instrument:
    """
    Instrument.
    """

    clefs: tuple[str, ...] = ()
    context: str = "Staff"
    # find_context_on_attach: typing.ClassVar[bool] = True
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[-inf, +inf]")

    check_effective_context: typing.ClassVar[bool] = True
    latent: typing.ClassVar[bool] = True
    persistent: typing.ClassVar[bool] = True
    redraw: typing.ClassVar[bool] = True
    site: typing.ClassVar[str] = "before"

    def __post_init__(self):
        assert isinstance(self.context, str), repr(self.context)
        assert isinstance(self.clefs, tuple), repr(self.clefs)
        assert all(isinstance(_, str) for _ in self.clefs)
        assert isinstance(self.pitch_range, _pcollections.PitchRange), repr(
            self.pitch_range
        )
        assert isinstance(self.middle_c_sounding_pitch, _pitch.NamedPitch), repr(
            self.middle_c_sounding_pitch
        )

    @property
    def _lilypond_type(self):
        if isinstance(self.context, type):
            return self.context.__name__
        elif isinstance(self.context, str):
            return self.context
        else:
            return type(self.context).__name__

    def _attachment_test_all(self, leaf):
        assert hasattr(leaf, "written_duration")
        if leaf._has_indicator(Instrument):
            string = f"Already has instrument: {leaf}."
            return string
        return True

    def _get_contributions(self, *, component=None, wrapper=None):
        contributions = _contributions.ContributionsBySite()
        site = getattr(contributions, self.site)
        strings = self._get_lilypond_format()
        assert isinstance(strings, list), repr(strings)
        site.commands.extend(strings)
        return contributions

    def _get_lilypond_format(self, context=None):
        return []


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class StringNumber:
    """
    String number.

    ..  container:: example

        String I:

        >>> abjad.StringNumber((1,))
        StringNumber(numbers=(1,))

        Strings II and III:

        >>> abjad.StringNumber((2, 3))
        StringNumber(numbers=(2, 3))

    """

    numbers: tuple[int, ...] = (1,)

    def __post_init__(self):
        assert isinstance(self.numbers, tuple), repr(self.numbers)
        assert all(0 < _ < 7 for _ in self.numbers)

    @property
    def roman_numerals(self) -> tuple[str, ...]:
        """
        Gets roman numerals of string number indicator.

        ..  container:: example

            String I:

            >>> indicator = abjad.StringNumber((1,))
            >>> indicator.roman_numerals
            ('i',)

            Strings II and III:

            >>> indicator = abjad.StringNumber((2, 3))
            >>> indicator.roman_numerals
            ('ii', 'iii')

        """
        numerals = ("i", "ii", "iii", "iv", "v", "vi")
        result = []
        for number in self.numbers:
            numeral = numerals[number - 1]
            result.append(numeral)
        return tuple(result)


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Tuning:
    """
    Tuning.

    ..  container:: example

        Violin tuning:

        >>> pitches = [abjad.NamedPitch(_) for _ in "G3 D4 A4 E5".split()]
        >>> abjad.Tuning(tuple(pitches))
        Tuning(pitches=(NamedPitch('g'), NamedPitch("d'"), NamedPitch("a'"), NamedPitch("e''")))

    """

    pitches: tuple[_pitch.NamedPitch, ...] = ()

    def __post_init__(self):
        assert isinstance(self.pitches, tuple), repr(self.pitches)
        assert all(isinstance(_, _pitch.NamedPitch) for _ in self.pitches)

    @property
    def pitch_ranges(self) -> list[_pcollections.PitchRange]:
        """
        Gets two-octave pitch-ranges for each pitch in this tuning.

        ..  container:: example

            >>> pitches = [abjad.NamedPitch(_) for _ in "G3 D4 A4 E5".split()]
            >>> tuning = abjad.Tuning(tuple(pitches))
            >>> for range_ in tuning.pitch_ranges:
            ...     range_
            PitchRange(range_string='[G3, G5]')
            PitchRange(range_string='[D4, D6]')
            PitchRange(range_string='[A4, A6]')
            PitchRange(range_string='[E5, E7]')

        """
        result = []
        for pitch in self.pitches or []:
            pitch_range: _pcollections.PitchRange = _pcollections.PitchRange(
                f"[{pitch.name}, {(pitch + 24).name}]"
            )
            result.append(pitch_range)
        return result

    def get_pitch_ranges_by_string_number(
        self, string_number: StringNumber
    ) -> tuple[_pcollections.PitchRange, ...]:
        """
        Gets tuning pitch ranges by string number.

        Violin tuning:

        ..  container:: example

            >>> pitches = [abjad.NamedPitch(_) for _ in "G3 D4 A4 E5".split()]
            >>> tuning = abjad.Tuning(tuple(pitches))
            >>> string_number = abjad.StringNumber((2, 3))
            >>> tuning.get_pitch_ranges_by_string_number(string_number)
            (PitchRange(range_string='[A4, A6]'), PitchRange(range_string='[D4, D6]'))

        """
        if not isinstance(string_number, StringNumber):
            string_number = StringNumber(string_number)
        assert isinstance(string_number, StringNumber)
        pitch_ranges = self.pitch_ranges
        result = []
        for number in string_number.numbers:
            index = -number
            pitch_range: _pcollections.PitchRange = pitch_ranges[index]
            result.append(pitch_range)
        return tuple(result)

    def get_pitches_by_string_number(
        self, string_number: StringNumber
    ) -> tuple[_pitch.NamedPitch, ...]:
        """
        Gets tuning pitches by string number.

        Violin tuning:

        ..  container:: example

            >>> pitches = [abjad.NamedPitch(_) for _ in "G3 D4 A4 E5".split()]
            >>> tuning = abjad.Tuning(tuple(pitches))
            >>> string_number = abjad.StringNumber((2, 3))
            >>> tuning.get_pitches_by_string_number(string_number)
            (NamedPitch("a'"), NamedPitch("d'"))

        """
        if not isinstance(string_number, StringNumber):
            string_number = StringNumber(string_number)
        assert isinstance(string_number, StringNumber)
        assert self.pitches is not None
        result = []
        for number in string_number.numbers:
            index = -number
            pitch = self.pitches[index]
            result.append(pitch)
        return tuple(result)

    def voice_pitch_classes(
        self,
        pitch_classes: list[_pitch.NamedPitchClass],
        allow_open_strings: bool = True,
    ) -> list[tuple[_pitch.NamedPitch | None, ...]]:
        r"""
        Voices ``pitch_classes``.

        ..  container:: example

            >>> pitches = [abjad.NamedPitch(_) for _ in "G3 D4 A4 E5".split()]
            >>> tuning = abjad.Tuning(tuple(pitches))
            >>> voicings = tuning.voice_pitch_classes([abjad.NamedPitchClass('a')])
            >>> for voicing in voicings:
            ...     voicing
            ...
            (NamedPitch('a'), None, None, None)
            (None, None, None, NamedPitch("a''"))
            (None, None, None, NamedPitch("a'''"))
            (None, None, NamedPitch("a'"), None)
            (None, None, NamedPitch("a''"), None)
            (None, None, NamedPitch("a'''"), None)
            (None, NamedPitch("a'"), None, None)
            (None, NamedPitch("a''"), None, None)
            (NamedPitch("a'"), None, None, None)

            >>> pcs = [abjad.NamedPitchClass(_) for _ in ["a", "d"]]
            >>> voicings = tuning.voice_pitch_classes(pcs, allow_open_strings=False)
            >>> for voicing in voicings:
            ...     voicing
            ...
            (NamedPitch('a'), None, None, NamedPitch("d'''"))
            (NamedPitch('a'), None, None, NamedPitch("d''''"))
            (NamedPitch('a'), None, NamedPitch("d''"), None)
            (NamedPitch('a'), None, NamedPitch("d'''"), None)
            (NamedPitch('a'), NamedPitch("d''"), None, None)
            (NamedPitch('a'), NamedPitch("d'''"), None, None)
            (None, None, NamedPitch("d''"), NamedPitch("a''"))
            (None, None, NamedPitch("d''"), NamedPitch("a'''"))
            (None, None, NamedPitch("a''"), NamedPitch("d'''"))
            (None, None, NamedPitch("a''"), NamedPitch("d''''"))
            (None, None, NamedPitch("d'''"), NamedPitch("a''"))
            (None, None, NamedPitch("d'''"), NamedPitch("a'''"))
            (None, None, NamedPitch("a'''"), NamedPitch("d'''"))
            (None, None, NamedPitch("a'''"), NamedPitch("d''''"))
            (None, NamedPitch("a'"), None, NamedPitch("d'''"))
            (None, NamedPitch("a'"), None, NamedPitch("d''''"))
            (None, NamedPitch("a'"), NamedPitch("d''"), None)
            (None, NamedPitch("a'"), NamedPitch("d'''"), None)
            (None, NamedPitch("d''"), None, NamedPitch("a''"))
            (None, NamedPitch("d''"), None, NamedPitch("a'''"))
            (None, NamedPitch("d''"), NamedPitch("a''"), None)
            (None, NamedPitch("d''"), NamedPitch("a'''"), None)
            (None, NamedPitch("a''"), None, NamedPitch("d'''"))
            (None, NamedPitch("a''"), None, NamedPitch("d''''"))
            (None, NamedPitch("a''"), NamedPitch("d''"), None)
            (None, NamedPitch("a''"), NamedPitch("d'''"), None)
            (None, NamedPitch("d'''"), None, NamedPitch("a''"))
            (None, NamedPitch("d'''"), None, NamedPitch("a'''"))
            (None, NamedPitch("d'''"), NamedPitch("a''"), None)
            (None, NamedPitch("d'''"), NamedPitch("a'''"), None)
            (NamedPitch("d'"), None, None, NamedPitch("a''"))
            (NamedPitch("d'"), None, None, NamedPitch("a'''"))
            (NamedPitch("d'"), None, NamedPitch("a''"), None)
            (NamedPitch("d'"), None, NamedPitch("a'''"), None)
            (NamedPitch("d'"), NamedPitch("a'"), None, None)
            (NamedPitch("d'"), NamedPitch("a''"), None, None)
            (NamedPitch("a'"), None, None, NamedPitch("d'''"))
            (NamedPitch("a'"), None, None, NamedPitch("d''''"))
            (NamedPitch("a'"), None, NamedPitch("d''"), None)
            (NamedPitch("a'"), None, NamedPitch("d'''"), None)
            (NamedPitch("a'"), NamedPitch("d''"), None, None)
            (NamedPitch("a'"), NamedPitch("d'''"), None, None)
            (NamedPitch("d''"), None, None, NamedPitch("a''"))
            (NamedPitch("d''"), None, None, NamedPitch("a'''"))
            (NamedPitch("d''"), None, NamedPitch("a''"), None)
            (NamedPitch("d''"), None, NamedPitch("a'''"), None)
            (NamedPitch("d''"), NamedPitch("a'"), None, None)
            (NamedPitch("d''"), NamedPitch("a''"), None, None)

        """
        assert self.pitches is not None
        assert all(isinstance(_, _pitch.NamedPitchClass) for _ in pitch_classes), repr(
            pitch_classes
        )
        nones = [None] * (len(self.pitches) - len(pitch_classes))
        pcs_and_nones = pitch_classes + nones
        permutations = _enumerate.yield_permutations(pcs_and_nones)
        unique_tuples = set([tuple(_) for _ in permutations])
        pitch_ranges = self.pitch_ranges
        result: list[tuple[_pitch.NamedPitch | None, ...]] = []
        for permutation in unique_tuples:
            sequences: list = []
            for pitch_range, pitch_class in zip(pitch_ranges, permutation):
                pitches: list[_pitch.NamedPitch | None]
                if pitch_class is None:
                    sequences.append([None])
                    continue
                pitches = list(pitch_range.voice_pitch_class(pitch_class))
                if not allow_open_strings:
                    pitches = [
                        pitch for pitch in pitches if pitch != pitch_range.start_pitch
                    ]
                if not pitches:
                    pitches = [None]
                sequences.append(pitches)
            subresult = _enumerate.outer_product(sequences)
            subresult = [tuple(x) for x in subresult]
            result.extend(subresult)
        assert isinstance(result, list)
        assert all(isinstance(_, tuple) for _ in result)
        result.sort()
        return result


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Accordion(Instrument):
    clefs: tuple[str, ...] = ("treble", "bass")
    context: str = "StaffGroup"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[E1, C8]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class AltoFlute(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("G3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[G3, G6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class AltoSaxophone(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Eb3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Db3, A5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class AltoTrombone(Instrument):
    clefs: tuple[str, ...] = ("bass", "tenor")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[A2, Bb5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class AltoVoice(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[F3, G5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class BaritoneSaxophone(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Eb2")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C2, Ab4]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class BaritoneVoice(Instrument):
    clefs: tuple[str, ...] = ("bass",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[A2, A4]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class BassClarinet(Instrument):
    clefs: tuple[str, ...] = ("treble", "bass")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Bb2")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Bb1, G5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class BassFlute(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C3, C6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class BassSaxophone(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Bb1")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Ab2, E4]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class BassTrombone(Instrument):
    clefs: tuple[str, ...] = ("bass",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C2, F4]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class BassVoice(Instrument):
    clefs: tuple[str, ...] = ("bass",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[E2, F4]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Bassoon(Instrument):
    clefs: tuple[str, ...] = ("bass", "tenor")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Bb1, Eb5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Cello(Instrument):
    clefs: tuple[str, ...] = ("bass", "tenor", "treble")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C2, G5]")
    tuning: Tuning = Tuning(
        tuple([_pitch.NamedPitch(_) for _ in "C2 G2 D3 A3".split()])
    )


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class ClarinetInA(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("A3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Db3, A6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class ClarinetInBFlat(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Bb3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[D3, Bb6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class ClarinetInEFlat(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Eb4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[F3, C7]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Contrabass(Instrument):
    clefs: tuple[str, ...] = ("bass", "treble")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C1, G4]")
    tuning: Tuning = Tuning(
        tuple([_pitch.NamedPitch(_) for _ in "C1 A1 D2 G2".split()])
    )


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class ContrabassClarinet(Instrument):
    clefs: tuple[str, ...] = ("treble", "bass")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Bb1")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Bb0, G4]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class ContrabassFlute(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("G2")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[G2, G5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class ContrabassSaxophone(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Eb1")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C1, Ab3]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Contrabassoon(Instrument):
    clefs: tuple[str, ...] = ("bass",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Bb0, Bb4]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class EnglishHorn(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("F3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[E3, C6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Flute(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C4, D7]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class FrenchHorn(Instrument):
    clefs: tuple[str, ...] = ("bass", "treble")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("F3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[B1, F5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Glockenspiel(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C6")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[G5, C8]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Guitar(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[E2, E5]")
    tuning: Tuning = Tuning(
        tuple([_pitch.NamedPitch(_) for _ in "E2 A2 D3 G3 B3 E4".split()])
    )


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Harp(Instrument):
    clefs: tuple[str, ...] = ("treble", "bass")
    context: str = "StaffGroup"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[B0, G#7]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Harpsichord(Instrument):
    clefs: tuple[str, ...] = ("treble", "bass")
    context: str = "StaffGroup"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C2, C7]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Marimba(Instrument):
    clefs: tuple[str, ...] = ("treble", "bass")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[F2, C7]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class MezzoSopranoVoice(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[A3, C6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Oboe(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Bb3, A6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Percussion(Instrument):
    clefs: tuple[str, ...] = ("percussion",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[-inf, +inf]")

    known_percussion = tuple(
        sorted(
            set(
                [
                    "agogô",
                    "anvil",
                    "bass drum",
                    "bongo drums",
                    "cabasa",
                    "cajón",
                    "castanets",
                    "caxixi",
                    "claves",
                    "conga drums",
                    "cowbell",
                    "crotales",
                    "cuíca",
                    "djembe",
                    "finger cymbals",
                    "flexatone",
                    "frame drum",
                    "gong",
                    "güiro",
                    "hand-held stones",
                    "jawbone",
                    "maracas",
                    "ratchet",
                    "rattle",
                    "sand blocks",
                    "scraped slate",
                    "siren",
                    "slapstick",
                    "slide whistle",
                    "snare drum",
                    "sponges",
                    "suspended cymbal",
                    "steel drums",
                    "tam-tam",
                    "tambourine",
                    "temple blocks",
                    "thunder machine",
                    "thundersheet",
                    "toms",
                    "tubular bells",
                    "triangle",
                    "vibraslap",
                    "whistle",
                    "wind chime",
                    "wind machine",
                    "wood blocks",
                    "wood planks",
                ]
            )
        )
    )


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Piano(Instrument):
    clefs: tuple[str, ...] = ("treble", "bass")
    context: str = "StaffGroup"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[A0, C8]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Piccolo(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C5")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[D5, C8]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class SopraninoSaxophone(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Eb4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Db4, F#6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class SopranoSaxophone(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Bb3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Ab3, E6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class SopranoVoice(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C4, E6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class TenorSaxophone(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Bb2")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Ab2, E5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class TenorTrombone(Instrument):
    clefs: tuple[str, ...] = ("tenor", "bass")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[E2, Eb5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class TenorVoice(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C3, D5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Trumpet(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[F#3, D6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Tuba(Instrument):
    clefs: tuple[str, ...] = ("bass",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[D1, F4]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Vibraphone(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[F3, F6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Viola(Instrument):
    clefs: tuple[str, ...] = ("alto", "treble")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C3, D6]")
    tuning: Tuning = Tuning(
        tuple([_pitch.NamedPitch(_) for _ in "C3 G3 D4 A4".split()])
    )


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Violin(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[G3, G7]")
    tuning: Tuning = Tuning(
        tuple([_pitch.NamedPitch(_) for _ in "G3 D4 A4 E5".split()])
    )


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Xylophone(Instrument):
    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C5")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C4, C7]")



import copy

import abjad


def test_Voice___copy___01():
    """
    Voices copy name, engraver removals, engraver consists, grob overrides and
    context settings. Voices do not copy components.
    """

    voice_1 = abjad.Voice("c'8 d'8 e'8 f'8")
    voice_1.name = "SopranoVoice"
    voice_1.remove_commands.append("Forbid_line_break_engraver")
    voice_1.consists_commands.append("Time_signature_engraver")
    abjad.override(voice_1).NoteHead.color = "#red"
    abjad.setting(voice_1).tupletFullLength = True
    voice_2 = copy.copy(voice_1)

    assert abjad.lilypond(voice_2) == abjad.string.normalize(
        r"""
        \context Voice = "SopranoVoice"
        \with
        {
            \remove Forbid_line_break_engraver
            \consists Time_signature_engraver
            \override NoteHead.color = #red
            tupletFullLength = ##t
        }
        {
        }
        """
    )


def test_Voice___delitem___01():
    """
    Delete container from voice.
    """

    voice = abjad.Voice(
        r"""
        c'8
        [
        \glissando
        {
            d'8
            \glissando
            e'8
            \glissando
        }
        f'8 ]
        """
    )

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            \glissando
            {
                d'8
                \glissando
                e'8
                \glissando
            }
            f'8
            ]
        }
        """
    )

    container = voice[1]
    del voice[1:2]

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            \glissando
            f'8
            ]
        }
        """
    )

    assert abjad.wf.wellformed(voice)
    assert abjad.wf.wellformed(container)


def test_Voice___len___01():
    """
    Voice length returns the number of elements in voice.
    """

    voice = abjad.Voice()
    assert len(voice) == 0


def test_Voice___len___02():
    """
    Voice length returns the number of elements in voice.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    assert len(voice) == 4


def test_Voice_lilypond_voice_resolution_01():
    """
    Anonymous voice with a sequence of leaves,
    in the middle of which there is a simultaneous,
    which in turn contains two anonymous voices.
    How does LilyPond resolve voices?
    LilyPond identifies three separate voices.
    LilyPond colors the outer four notes (c'8 d'8 b'8 c''8) red.
    LilyPond colors the inner four notes black.
    LilyPond issues clashing note column warnings for the inner notes.
    How should Abjad resolve voices?
    """

    voice = abjad.Voice("c'8 d'8 b'8 c''8")
    voice.insert(2, abjad.Container([abjad.Voice("e'8 f'8"), abjad.Voice("g'8 a'8")]))
    voice[2].simultaneous = True
    abjad.override(voice).NoteHead.color = "#red"

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        \with
        {
            \override NoteHead.color = #red
        }
        {
            c'8
            d'8
            <<
                \new Voice
                {
                    e'8
                    f'8
                }
                \new Voice
                {
                    g'8
                    a'8
                }
            >>
            b'8
            c''8
        }
        """
    )


def test_Voice_lilypond_voice_resolution_02():
    """
    Named voice with  with a sequence of leaves,
    in the middle of which there is a simultaneous,
    which in turn contains one like-named and one differently named voice.
    How does LilyPond resolve voices?
    """

    voice = abjad.Voice("c'8 d'8 b'8 c''8")
    voice.name = "foo"
    voice.insert(2, abjad.Container([abjad.Voice("e'8 f'8"), abjad.Voice("g'8 a'8")]))
    voice[2].simultaneous = True
    voice[2][0].name = "foo"
    abjad.override(voice).NoteHead.color = "#red"

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \context Voice = "foo"
        \with
        {
            \override NoteHead.color = #red
        }
        {
            c'8
            d'8
            <<
                \context Voice = "foo"
                {
                    e'8
                    f'8
                }
                \new Voice
                {
                    g'8
                    a'8
                }
            >>
            b'8
            c''8
        }
        """
    )

    """
    LilyPond colors six notes red and two notes black.
    LilyPond identifies two voices.
    """


def test_Voice_lilypond_voice_resolution_03():
    """
    Container containing a run of leaves.
    Two like-structured simultaneouss in the middle of the run.
    LilyPond handles this example perfectly.
    LilyPond colors the four note_heads of the soprano voice red.
    LilyPond colors all other note_heads black.
    """

    container = abjad.Container(
        r"""
        c'8
        <<
            \context Voice = "alto" {
                d'8
                e'8
            }
            \context Voice = "soprano" {
                f'8
                g'8
            }
        >>
        <<
            \context Voice = "alto" {
                a'8
                b'8
            }
            \context Voice = "soprano" {
                c''8
                d''8
            }
        >>
        e''8
        """
    )

    abjad.override(container[1][1]).NoteHead.color = "#red"
    abjad.override(container[2][1]).NoteHead.color = "#red"

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            c'8
            <<
                \context Voice = "alto"
                {
                    d'8
                    e'8
                }
                \context Voice = "soprano"
                \with
                {
                    \override NoteHead.color = #red
                }
                {
                    f'8
                    g'8
                }
            >>
            <<
                \context Voice = "alto"
                {
                    a'8
                    b'8
                }
                \context Voice = "soprano"
                \with
                {
                    \override NoteHead.color = #red
                }
                {
                    c''8
                    d''8
                }
            >>
            e''8
        }
        """
    )
